# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
from __future__ import annotations

import builtins as __builtin__
import gc
from logging import getLogger
import sys
from typing import Any

import pytest

from duecredit.collector import DueCreditCollector
from duecredit.entries import Doi
import duecredit.tests.mod as mod

from .. import __version__
from ..injections.injector import (
    DueCreditInjector,
    find_object,
    get_modules_for_injection,
)
from ..versions import Version

try:
    import mvpa2  # noqa: F401

    _have_mvpa2 = True
except ImportError:
    _have_mvpa2 = False

_orig__import__ = __builtin__.__import__
lgr = getLogger("duecredit.tests.injector")


class TestActiveInjector:
    def setup_method(self) -> None:
        lgr.log(5, "Setting up for a TestActiveInjector test")
        self._cleanup_modules()
        self.due = DueCreditCollector()
        self.injector = DueCreditInjector(collector=self.due)
        self.injector.activate(retrospect=False)  # numpy might be already loaded...

    def teardown_method(self) -> None:
        lgr.log(5, "Tearing down after a TestActiveInjector test")
        # gc might not pick up inj after some tests complete
        # so we will always deactivate explicitly
        self.injector.deactivate()
        assert __builtin__.__import__ is _orig__import__
        self._cleanup_modules()

    def _cleanup_modules(self) -> None:
        if "duecredit.tests.mod" in sys.modules:
            sys.modules.pop("duecredit.tests.mod")

    def _test_simple_injection(self, func, import_stmt, func_call=None) -> None:
        assert "duecredit.tests.mod" not in sys.modules
        self.injector.add(
            "duecredit.tests.mod",
            func,
            Doi("1.2.3.4"),
            description="Testing %s" % func,
            min_version="0.1",
            max_version="1.0",
            tags=["implementation", "very custom"],
        )
        assert "duecredit.tests.mod" not in sys.modules  # no import happening
        assert len(self.due._entries) == 0
        assert len(self.due.citations) == 0

        globals_: dict[str, Any] = {}
        locals_: dict[str, Any] = {}
        exec(import_stmt, globals_, locals_)

        assert len(self.due._entries) == 1  # we should get an entry now
        assert len(self.due.citations) == 0  # but not yet a citation

        import duecredit.tests.mod as mod

        _, _, obj = find_object(mod, func)
        assert obj.__duecredited__  # we wrapped
        assert obj.__duecredited__ is not obj  # and it is not pointing to the same func
        assert obj.__doc__ == "custom docstring"  # we preserved docstring

        # TODO: test decoration features -- preserver __doc__ etc
        exec('ret = %s(None, "somevalue")' % (func_call or func), globals_, locals_)
        # TODO: awkwardly 'ret' is not found in the scope while running pytest
        # under python3.4, although present in locals()... WTF?
        assert locals_["ret"] == "%s: None, somevalue" % func
        assert len(self.due._entries) == 1
        assert len(self.due.citations) == 1

        # TODO: there must be a cleaner way to get first value
        citation = list(self.due.citations.values())[0]
        # TODO: ATM we don't allow versioning of the submodules -- we should
        # assert_equal(citation.version, '0.5')
        # ATM it will be the duecredit's version
        assert citation.version == Version(__version__)

        assert citation.tags == ["implementation", "very custom"]

    def _test_double_injection(self, func, import_stmt, func_call=None) -> None:
        assert "duecredit.tests.mod" not in sys.modules
        # add one injection
        self.injector.add(
            "duecredit.tests.mod",
            func,
            Doi("1.2.3.4"),
            description="Testing %s" % func,
            min_version="0.1",
            max_version="1.0",
            tags=["implementation", "very custom"],
        )
        # add another one
        self.injector.add(
            "duecredit.tests.mod",
            func,
            Doi("1.2.3.5"),
            description="Testing %s" % func,
            min_version="0.1",
            max_version="1.0",
            tags=["implementation", "very custom"],
        )
        assert "duecredit.tests.mod" not in sys.modules  # no import happening
        assert len(self.due._entries) == 0
        assert len(self.due.citations) == 0

        globals_: dict[str, Any] = {}
        locals_: dict[str, Any] = {}
        exec(import_stmt, globals_, locals_)

        assert len(self.due._entries) == 2  # we should get two entries now
        assert len(self.due.citations) == 0  # but not yet a citation

        import duecredit.tests.mod as mod

        _, _, obj = find_object(mod, func)
        assert obj.__duecredited__  # we wrapped
        assert obj.__duecredited__ is not obj  # and it is not pointing to the same func
        assert obj.__doc__ == "custom docstring"  # we preserved docstring

        # TODO: test decoration features -- preserver __doc__ etc
        exec('ret = %s(None, "somevalue")' % (func_call or func), globals_, locals_)
        # TODO: awkwardly 'ret' is not found in the scope while running pytest
        # under python3.4, although present in locals()... WTF?
        assert locals_["ret"] == "%s: None, somevalue" % func
        assert len(self.due._entries) == 2
        assert len(self.due.citations) == 2

        # TODO: there must be a cleaner way to get first value
        citation = list(self.due.citations.values())[0]
        # TODO: ATM we don't allow versioning of the submodules -- we should
        # assert_equal(citation.version, '0.5')
        # ATM it will be the duecredit's version
        assert citation.version, __version__

        assert citation.tags == ["implementation", "very custom"]

    test1 = ("testfunc1", "from duecredit.tests.mod import testfunc1", None)
    test2 = (
        "TestClass1.testmeth1",
        "from duecredit.tests.mod import TestClass1; c = TestClass1()",
        "c.testmeth1",
    )
    test3 = (
        "TestClass12.Embed.testmeth1",
        "from duecredit.tests.mod import TestClass12; c = TestClass12.Embed()",
        "c.testmeth1",
    )

    @pytest.mark.parametrize("func, import_stmt, func_call", [test1, test2, test3])
    def test_simple_injection(self, func, import_stmt, func_call) -> None:
        self._test_simple_injection(func, import_stmt, func_call)

    @pytest.mark.parametrize("func, import_stmt, func_call", [test1, test2, test3])
    def test_double_injection(self, func, import_stmt, func_call) -> None:
        self._test_double_injection(func, import_stmt, func_call)

    def test_delayed_entries(self) -> None:
        # verify that addition of delayed injections happened
        modules_for_injection = get_modules_for_injection()
        assert len(self.injector._delayed_injections) == len(modules_for_injection)
        assert self.injector._entry_records == {}  # but no entries were added
        assert "scipy" in self.injector._delayed_injections  # We must have it ATM

        try:
            # We do have injections for scipy
            import scipy  # noqa: F401
        except ImportError as e:
            pytest.skip("scipy was not found: {}".format(e))

    def test_import_mvpa2_suite(self) -> None:
        if not _have_mvpa2:
            pytest.skip("no mvpa2 found")
        # just a smoke test for now
        import mvpa2.suite as mv  # noqa: F401

    def _test_incorrect_path(self, mod, obj) -> None:
        ref = Doi("1.2.3.4")
        # none of them should lead to a failure
        self.injector.add(mod, obj, ref)
        # now cause the import handling -- it must not fail
        # TODO: catch/analyze warnings
        exec("from duecredit.tests.mod import testfunc1", {}, {})

    @pytest.mark.parametrize(
        "mod, obj",
        [
            ("nonexistingmodule", None),
            ("duecredit.tests.mod.nonexistingmodule", None),
            ("duecredit.tests.mod", "nonexisting"),
            ("duecredit.tests.mod", "nonexisting.whocares"),
        ],
    )
    def test_incorrect_path(self, mod, obj):
        self._test_incorrect_path(mod, obj)


def test_find_iobject() -> None:
    assert find_object(mod, "testfunc1") == (mod, "testfunc1", mod.testfunc1)
    assert find_object(mod, "TestClass1") == (mod, "TestClass1", mod.TestClass1)
    assert find_object(mod, "TestClass1.testmeth1") == (
        mod.TestClass1,
        "testmeth1",
        mod.TestClass1.testmeth1,
    )
    assert find_object(mod, "TestClass12.Embed.testmeth1") == (
        mod.TestClass12.Embed,
        "testmeth1",
        mod.TestClass12.Embed.testmeth1,
    )


def test_no_double_activation() -> None:
    orig__import__ = __builtin__.__import__
    try:
        due = DueCreditCollector()
        injector = DueCreditInjector(collector=due)
        injector.activate()
        assert __builtin__.__import__ is not orig__import__
        duecredited__import__ = __builtin__.__import__
        # TODO: catch/analyze/swallow warning
        injector.activate()
        assert (
            __builtin__.__import__ is duecredited__import__
        )  # we didn't decorate again
    finally:
        injector.deactivate()
        __builtin__.__import__ = orig__import__


def test_get_modules_for_injection() -> None:
    # output order is sorted by name (not that it matters for functionality)
    assert get_modules_for_injection() == [
        "mod_biosig",
        "mod_dipy",
        "mod_matplotlib",
        "mod_mdp",
        "mod_mne",
        "mod_nibabel",
        "mod_nipy",
        "mod_nipype",
        "mod_numpy",
        "mod_pandas",
        "mod_psychopy",
        "mod_scipy",
        "mod_skimage",
        "mod_sklearn",
    ]


def test_cover_our_injections() -> None:
    # this one tests only import/syntax/api for the injections
    due = DueCreditCollector()
    inj = DueCreditInjector(collector=due)
    for modname in get_modules_for_injection():
        mod = __import__(
            "duecredit.injections." + modname, fromlist=["duecredit.injections"]
        )
        mod.inject(inj)


def test_no_harm_from_deactivate() -> None:
    # if we have not activated one -- shouldn't blow if we deactivate it
    # TODO: catch warning being spitted out
    DueCreditInjector().deactivate()


def test_injector_del() -> None:
    orig__import__ = __builtin__.__import__
    try:
        due = DueCreditCollector()
        inj = DueCreditInjector(collector=due)
        del inj  # delete inactive
        assert __builtin__.__import__ is orig__import__
        inj = DueCreditInjector(collector=due)
        inj.activate(retrospect=False)
        assert __builtin__.__import__ is not orig__import__
        assert inj._orig_import is not None
        del inj  # delete active but not used
        inj = None  # type: ignore  # noqa: F841
        __builtin__.__import__ = None  # type: ignore
        # /\ We need to do that since otherwise gc will not pick up inj
        gc.collect()  # To cause __del__
        assert __builtin__.__import__ is orig__import__
        import abc  # noqa: F401   # and new imports work just fine
    finally:
        __builtin__.__import__ = orig__import__


def test_injector_delayed_del() -> None:
    # interesting case -- if we still have an instance of injector hanging around
    # and then create a new one, activate it but then finally delete/gc old one
    # it would (currently) reset import back (because atm defined as class var)
    # which would ruin operation of the new injector
    orig__import__ = __builtin__.__import__
    try:
        due = DueCreditCollector()

        inj = DueCreditInjector(collector=due)
        inj.activate(retrospect=False)
        assert __builtin__.__import__ is not orig__import__
        assert inj._orig_import is not None
        inj.deactivate()
        assert __builtin__.__import__ is orig__import__
        assert inj._orig_import is None

        # create 2nd one
        inj2 = DueCreditInjector(collector=due)
        inj2.activate(retrospect=False)
        assert __builtin__.__import__ is not orig__import__
        assert inj2._orig_import is not None
        del inj  # type: ignore
        inj = None  # type: ignore  # noqa: F841
        gc.collect()  # To cause __del__
        assert (
            __builtin__.__import__ is not orig__import__
        )  # would fail if del had side-effect
        assert inj2._orig_import is not None
        inj2.deactivate()
        assert __builtin__.__import__ is orig__import__
        import abc  # noqa: F401   # and new imports work just fine
    finally:
        __builtin__.__import__ = orig__import__



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Importer which would also call decoration on a module upon import
"""
from __future__ import annotations

__docformat__ = "restructuredtext"

import builtins as __builtin__
from functools import wraps
from glob import glob
import logging
import os
from os.path import basename, dirname
from os.path import join as pathjoin
import sys
from typing import TYPE_CHECKING, Any

from ..log import lgr

if TYPE_CHECKING:
    from ..entries import BibTeX, Doi, Url

__all__ = ["DueCreditInjector", "find_object"]


# TODO: move elsewhere
def _short_str(obj: Any, lng: int = 30) -> str:
    """Return a shortened str of an object -- for logging"""
    s = str(obj)
    if len(s) > lng:
        return s[: lng - 3] + "..."
    else:
        return s


def get_modules_for_injection() -> list[str]:
    """Get local modules which provide "inject" method to provide delayed population of injector"""
    return sorted(
        basename(x)[:-3] for x in glob(pathjoin(dirname(__file__), "mod_*.py"))
    )


def find_object(mod: Any, path: str) -> tuple[Any, str, Any]:
    """Finds object among present within module "mod" given path specification within

    Returns
    -------

    parent, obj_name, obj
    """
    obj = mod  # we will look first within module
    for obj_name in path.split("."):
        parent = obj
        obj = getattr(parent, obj_name)
    return parent, obj_name, obj


# We will keep a very original __import__ to mitigate cases of buggy python
# behavior, see e.g.
# https://github.com/duecredit/duecredit/issues/40
# But we will also keep the __import__ as of 'activate' call state so we could
# stay friendly to anyone else who might decorate __import__ as well
_very_orig_import = __builtin__.__import__


class DueCreditInjector:
    """Takes care about "injecting" duecredit references into 3rd party modules upon their import

    First entries to be "injected" need to be add'ed to the instance.
    To not incur significant duecredit startup penalty, those entries are added
    for a corresponding package only whenever corresponding top-level module gets
    imported.
    """

    # Made as a class variable to assure being singleton and available upon del
    __orig_import = None

    # and interact with its value through the property to ease tracking of actions
    # performed on it
    @property
    def _orig_import(self):
        return DueCreditInjector.__orig_import

    @_orig_import.setter
    def _orig_import(self, value) -> None:
        lgr.log(
            2,
            "Reassigning _orig_import from %r to %r",
            DueCreditInjector.__orig_import,
            value,
        )
        DueCreditInjector.__orig_import = value

    def __init__(self, collector=None) -> None:
        if collector is None:
            from duecredit import due

            collector = due
        self._collector = collector
        self._delayed_injections: dict[str, str] = {}
        self._entry_records: dict[
            str, dict[str | None, Any]
        ] = {}  # dict:  modulename: {object: [('entry', cite kwargs)]}
        self._processed_modules: set[str] = set()
        # We need to process modules only after we are done with all nested imports, otherwise we
        # might be trying to process them too early -- whenever they are not yet linked to their
        # parent's namespace. So we will keep track of import level and set of modules which
        # would need to be processed whenever we are back at __import_level == 1
        self.__import_level = 0
        self.__queue_to_process: set[str] = set()
        self.__processing_queue = False
        self._active = False
        lgr.debug("Created injector %r", self)

    def _populate_delayed_injections(self) -> None:
        self._delayed_injections = {}
        for inj_mod_name in get_modules_for_injection():
            assert inj_mod_name.startswith("mod_")
            mod_name = inj_mod_name[4:]
            lgr.debug("Adding delayed injection for %s", (mod_name,))
            self._delayed_injections[mod_name] = inj_mod_name

    def add(
        self,
        modulename: str,
        obj: str | None,
        entry: Doi | BibTeX | Url,
        min_version=None,
        max_version=None,
        **kwargs: Any,
    ) -> None:
        """Add a citation for a given module or object within it

        Parameters
        ----------
        modulename : string
          Name of the module (possibly a sub-module)
        obj : string or None
          Name of the object (function, method within a class) or None (if for entire module)
        min_version, max_version : string or tuple, optional
          Min (inclusive) / Max (exclusive) version of the module where this
          citation is applicable
        **kwargs
          Keyword arguments to be passed into cite. Note that "path" will be automatically set
          if not provided
        """
        lgr.debug(
            "Adding citation entry %s for %s:%s", _short_str(entry), modulename, obj
        )
        if modulename not in self._entry_records:
            self._entry_records[modulename] = {}
        if obj not in self._entry_records[modulename]:
            self._entry_records[modulename][obj] = []
        obj_entries = self._entry_records[modulename][obj]
        if "path" not in kwargs:
            kwargs["path"] = modulename + ((":%s" % obj) if obj else "")
        obj_entries.append(
            {
                "entry": entry,
                "kwargs": kwargs,
                "min_version": min_version,
                "max_version": max_version,
            }
        )

    @property
    def _import_level_prefix(self) -> str:
        return "." * self.__import_level

    def _process_delayed_injection(self, mod_name: str) -> None:
        lgr.debug(
            "%sProcessing delayed injection for %s", self._import_level_prefix, mod_name
        )
        inj_mod_name = self._delayed_injections[mod_name]
        assert not hasattr(self._orig_import, "__duecredited__")
        try:
            inj_mod_name_full = "duecredit.injections." + inj_mod_name
            lgr.log(3, "Importing %s", inj_mod_name_full)
            # Mark it is a processed already, to avoid its processing etc
            self._processed_modules.add(inj_mod_name_full)
            inj_mod = self._orig_import(
                inj_mod_name_full, fromlist=["duecredit.injections"]
            )
        except Exception as e:
            if os.environ.get("DUECREDIT_ALLOW_FAIL", False):
                raise
            raise RuntimeError("Failed to import {}: {!r}".format(inj_mod_name, e))
        # TODO: process min/max_versions etc
        assert hasattr(inj_mod, "inject")
        lgr.log(3, "Calling injector of %s", inj_mod_name_full)
        inj_mod.inject(self)

    def process(self, mod_name: str) -> None:
        """Process import of the module, possibly decorating some methods with duecredit entries"""
        assert (
            self.__import_level == 0
        )  # we should never process while nested within imports
        # We need to mark that module as processed EARLY, so we don't try to re-process it
        # while doing _process_delayed_injection
        self._processed_modules.add(mod_name)

        if mod_name in self._delayed_injections:
            # should be hit only once, "theoretically" unless I guess reimport is used etc
            self._process_delayed_injection(mod_name)

        if mod_name not in self._entry_records:
            return

        total_number_of_citations = sum(
            map(len, self._entry_records[mod_name].values())
        )
        lgr.log(
            logging.DEBUG + 5,
            "Process %d citation injections for %d objects for module %s",
            total_number_of_citations,
            len(self._entry_records[mod_name]),
            mod_name,
        )

        try:
            mod = sys.modules[mod_name]
        except KeyError:
            lgr.warning("Failed to access module %s among sys.modules" % mod_name)
            return

        # go through the known entries and register them within the collector, and
        # decorate corresponding methods
        # There could be multiple records per module
        for obj_path, obj_entry_records in self._entry_records[mod_name].items():
            parent, obj_name = None, None
            if obj_path:
                # so we point to an object within the mod
                try:
                    parent, obj_name, obj = find_object(mod, obj_path)
                except (KeyError, AttributeError) as e:
                    lgr.warning(
                        "Could not find {} in module {}: {}".format(obj_path, mod, e)
                    )
                    continue

            # there could be multiple per func
            lgr.log(
                4,
                "Considering %d records for decoration of %s:%s",
                len(obj_entry_records),
                parent,
                obj_name,
            )
            for obj_entry_record in obj_entry_records:
                entry = obj_entry_record["entry"]
                # Add entry explicitly
                self._collector.add(entry)
                if obj_path:  # if not entire module -- decorate!
                    assert obj_name
                    decorator = self._collector.dcite(
                        entry.get_key(), **obj_entry_record["kwargs"]
                    )
                    lgr.debug("Decorating %s:%s with %s", parent, obj_name, decorator)
                    obj_decorated = decorator(obj)
                    setattr(parent, obj_name, obj_decorated)
                    # override previous obj with the decorated one if there are multiple decorators
                    obj = obj_decorated
                else:
                    lgr.log(
                        3,
                        "Citing directly %s:%s since obj_path is empty",
                        parent,
                        obj_name,
                    )
                    self._collector.cite(entry.get_key(), **obj_entry_record["kwargs"])

        lgr.log(3, "Done processing injections for module %s", mod_name)

    def _mitigate_None_orig_import(self, name: str, *args: Any, **kwargs: Any) -> Any:
        lgr.error(
            "For some reason self._orig_import is None"
            ". Importing using stock importer to mitigate and adjusting _orig_import"
        )
        self._orig_import = _very_orig_import
        return _very_orig_import(name, *args, **kwargs)

    def activate(self, retrospect=True):
        """
        Parameters
        ----------
        retrospect : bool, optional
          Either consider already loaded modules
        """

        if not self._orig_import:
            # for paranoid Yarik so we have assurance we are not somehow
            # overriding our decorator
            if hasattr(__builtin__.__import__, "__duecredited__"):
                raise RuntimeError("__import__ is already duecredited")

            self._orig_import = __builtin__.__import__

            @wraps(__builtin__.__import__)
            def __import(name, *args, **kwargs):
                if (
                    self.__processing_queue
                    or name in self._processed_modules
                    or name in self.__queue_to_process
                ):
                    lgr.debug("Performing undecorated import of %s", name)
                    # return right away without any decoration in such a case
                    if self._orig_import:
                        return _very_orig_import(name, *args, **kwargs)
                    else:
                        return self._mitigate_None_orig_import(name, *args, **kwargs)
                import_level_prefix = self._import_level_prefix
                lgr.log(
                    1, "%sProcessing request to import %s", import_level_prefix, name
                )
                # importing submodule might result in importing a new one and
                # name here is not sufficient to determine which module would actually
                # get imported unless level=0 (absolute import), but that one rarely used

                # could be old-style or new style relative import!
                # args[0] -> globals, [1] -> locals(), [2] -> fromlist, [3] -> level
                level = args[3] if len(args) >= 4 else kwargs.get("level", -1)
                # fromlist = args[2] if len(args) >= 3 else kwargs.get('fromlist', [])

                if not retrospect and not self._processed_modules:
                    # we were asked to not consider those modules which were already loaded
                    # so let's assume that they were all processed already
                    self._processed_modules = set(sys.modules)

                mod = None
                try:
                    self.__import_level += 1
                    # TODO: safe-guard all our logic so
                    # if anything goes wrong post-import -- we still return imported module
                    if self._orig_import:
                        mod = self._orig_import(name, *args, **kwargs)
                    else:
                        mod = self._mitigate_None_orig_import(name, *args, **kwargs)

                    self._handle_fresh_imports(name, import_level_prefix, level)
                finally:
                    self.__import_level -= 1

                if self.__import_level == 0 and self.__queue_to_process:
                    self._process_queue()

                lgr.log(1, "%sReturning %s", import_level_prefix, mod)
                return mod

            __import.__duecredited__ = True

            self._populate_delayed_injections()

            if retrospect:
                lgr.debug("Considering previously loaded %d modules", len(sys.modules))
                # operate on keys() (not iterator) since we might end up importing delayed injection modules etc
                for mod_name in sys.modules.keys():
                    self.process(sys.modules[mod_name])

            lgr.debug("Assigning our importer")
            __builtin__.__import__ = __import
            self._active = True

        else:
            lgr.warning(
                "Seems that we are calling duecredit_importer twice."
                " No harm is done but shouldn't happen"
            )

    def _handle_fresh_imports(self, name: str, import_level_prefix: str, level) -> None:
        """Check which modules were imported since last point we checked and add them to the queue"""
        new_imported_modules = (
            set(sys.modules.keys()) - self._processed_modules - self.__queue_to_process
        )
        if new_imported_modules:
            lgr.log(
                4,
                "%s%d new modules were detected upon import of %s (level=%s)",
                import_level_prefix,
                len(new_imported_modules),
                name,
                level,
            )
            # lgr.log(2, "%s%d new modules were detected: %s, upon import of %s (level=%s)",
            #        import_level_prefix, len(new_imported_modules), new_imported_modules, name, level)
        for imported_mod in new_imported_modules:
            if imported_mod in self.__queue_to_process:
                # we saw it already
                continue
            # lgr.log(1, "Name %r was imported as %r (path: %s). fromlist: %s, level: %s",
            #        name, mod.__name__, getattr(mod, '__path__', None), fromlist, level)
            # package
            package = imported_mod.split(".", 1)[0]
            if (
                package != imported_mod
                and package not in self._processed_modules
                and package not in self.__queue_to_process
            ):
                # if its parent package wasn't yet imported before
                lgr.log(
                    3,
                    "%sParent of %s, %s wasn't yet processed, adding to the queue",
                    import_level_prefix,
                    imported_mod,
                    package,
                )
                self.__queue_to_process.add(package)
            self.__queue_to_process.add(imported_mod)

    def _process_queue(self) -> None:
        """Process the queue of collected imported modules"""
        # process the queue
        lgr.debug(
            "Processing queue of imported %d modules", len(self.__queue_to_process)
        )
        # We need first to process top-level modules etc, so delayed injections get picked up,
        # let's sort by the level
        queue_with_levels = sorted((m.count("."), m) for m in self.__queue_to_process)
        self.__processing_queue = True
        try:
            sorted_queue = [x[1] for x in queue_with_levels]
            while sorted_queue:
                mod_name = sorted_queue.pop(0)
                self.process(mod_name)
                self.__queue_to_process.remove(mod_name)
            assert not len(self.__queue_to_process)
        finally:
            self.__processing_queue = False

    def deactivate(self) -> None:
        if not self._orig_import:
            lgr.warning(
                "_orig_import is not yet known, so we haven't decorated default importer yet."
                " Nothing TODO"
            )
            return
        if not self._active:  # pragma: no cover
            lgr.error("Must have not happened, but we will survive!")
        lgr.debug("Assigning original importer")
        __builtin__.__import__ = self._orig_import
        self._orig_import = None
        self._active = False

    def __del__(self) -> None:
        if lgr:
            lgr.debug("%s is asked to be deleted", self)
        try:
            if self._active:
                self.deactivate()
        except:  # noqa: E722, B001
            pass
        try:
            super(self.__class__, self).__del__()  # type: ignore
        except:  # noqa: E722, B001
            pass



# ex: set sts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
from __future__ import annotations

from io import StringIO
import os
import pickle
import random
import re
from typing import Any

import pytest
from pytest import MonkeyPatch

import duecredit.io

from .test_collector import _sample_bibtex, _sample_bibtex2, _sample_doi
from ..collector import CitationKey, DueCreditCollector
from ..entries import BibTeX, Doi, Text, Url
from ..io import (
    BibTeXOutput,
    Output,
    PickleOutput,
    TextOutput,
    _is_contained,
    format_bibtex,
    get_text_rendering,
    import_doi,
)

try:
    # TODO: for some reason test below started to complain that we are trying
    # to overwrite the cassette.
    #
    # import vcr

    # @vcr.use_cassette()
    def test_import_doi() -> None:
        doi_good = "10.1038/nrd842"
        assert isinstance(import_doi(doi_good, sleep=0.00001, retries=2), str)

        doi_bad = "fasljfdldaksj"
        with pytest.raises(ValueError):
            import_doi(doi_bad, sleep=0.00001, retries=2)

        doi_zenodo = "10.5281/zenodo.50186"
        assert isinstance(import_doi(doi_zenodo, sleep=0.00001, retries=2), str)

except ImportError:
    # no vcr, and that is in 2015!
    pass


def test_pickleoutput(tmpdir) -> None:
    # entry = BibTeX('@article{XXX0, ...}')
    entry = BibTeX(
        "@article{Atkins_2002,\n"
        "title=title,\n"
        "volume=1, \n"
        "url=https://doi.org/10.1038/nrd842, \n"
        "DOI=10.1038/nrd842, \n"
        "number=7, \n"
        "journal={Nat. Rev. Drug Disc.}, \n"
        "publisher={Nature Publishing Group}, \n"
        "author={Atkins, Joshua H. and Gershell, Leland J.}, \n"
        "year={2002}, \n"
        "month={Jul}, \n"
        "pages={491--492}\n}"
    )
    collector_ = DueCreditCollector()
    collector_.add(entry)
    collector_.cite(entry, path="module")

    # test it doesn't puke with an empty collector
    collectors = [collector_, DueCreditCollector()]

    tempfile = str(tmpdir.mkdir("sub").join("tempfile.txt"))

    for collector in collectors:
        pickler = PickleOutput(collector, fn=tempfile)
        assert pickler.fn == tempfile
        assert pickler.dump() is None  # type: ignore

        with open(tempfile, "rb") as f:
            collector_loaded = pickle.load(f)

        assert collector.citations.keys() == collector_loaded.citations.keys()
        # TODO: implement comparison of citations
        assert collector._entries.keys() == collector_loaded._entries.keys()
        os.unlink(tempfile)


def test_output() -> None:
    entry = BibTeX(_sample_bibtex)
    entry2 = BibTeX(_sample_bibtex2)

    # normal use
    collector = DueCreditCollector()
    collector.cite(entry, path="package")
    collector.cite(entry, path="package.module")

    output = Output(None, collector)

    packages, modules, objects = output._get_collated_citations(tags=["*"])

    assert len(packages) == 1
    assert len(modules) == 1
    assert len(objects) == 0

    assert (
        packages["package"][0]
        == collector.citations[CitationKey("package", entry.get_key())]
    )
    assert (
        modules["package.module"][0]
        == collector.citations[CitationKey("package.module", entry.get_key())]
    )

    # no toppackage
    collector = DueCreditCollector()
    collector.cite(entry, path="package")
    collector.cite(entry, path="package2.module")

    output = Output(None, collector)

    packages, modules, objects = output._get_collated_citations(tags=["*"])

    assert len(packages) == 0
    assert len(modules) == 1
    assert len(objects) == 0

    assert (
        modules["package2.module"][0]
        == collector.citations[CitationKey("package2.module", entry.get_key())]
    )

    # toppackage because required
    collector = DueCreditCollector()
    collector.cite(entry, path="package", cite_module=True)
    collector.cite(entry, path="package2.module")

    output = Output(None, collector)

    packages, modules, objects = output._get_collated_citations(tags=["*"])

    assert len(packages) == 1
    assert len(modules) == 1
    assert len(objects) == 0

    assert (
        packages["package"][0]
        == collector.citations[CitationKey("package", entry.get_key())]
    )
    assert (
        modules["package2.module"][0]
        == collector.citations[CitationKey("package2.module", entry.get_key())]
    )

    # check it returns multiple entries
    collector = DueCreditCollector()
    collector.cite(entry, path="package")
    collector.cite(entry2, path="package")
    collector.cite(entry, path="package.module")

    output = Output(None, collector)

    packages, modules, objects = output._get_collated_citations(tags=["*"])

    assert len(packages) == 1
    assert len(packages["package"]) == 2
    assert len(modules) == 1
    assert len(objects) == 0

    # sort them in order so we know who is who
    # entry2 key is Atk...
    # entry key is XX..
    packs = sorted(packages["package"], key=lambda x: x.entry.key)

    assert packs[0] == collector.citations[CitationKey("package", entry2.get_key())]
    assert packs[1] == collector.citations[CitationKey("package", entry.get_key())]
    assert (
        modules["package.module"][0]
        == collector.citations[CitationKey("package.module", entry.get_key())]
    )

    # check that filtering works
    collector = DueCreditCollector()
    collector.cite(entry, path="package", tags=["edu"])
    collector.cite(entry2, path="package")
    collector.cite(entry, path="package.module", tags=["edu"])

    output = Output(None, collector)

    packages, modules, objects = output._get_collated_citations(tags=["edu"])

    assert len(packages) == 1
    assert len(packages["package"]) == 1
    assert len(modules) == 1
    assert len(objects) == 0

    assert (
        packages["package"][0]
        == collector.citations[CitationKey("package", entry.get_key())]
    )
    assert (
        modules["package.module"][0]
        == collector.citations[CitationKey("package.module", entry.get_key())]
    )


def test_output_return_all(monkeypatch: MonkeyPatch) -> None:
    entry = BibTeX(_sample_bibtex)
    entry2 = BibTeX(_sample_bibtex2)

    # normal use
    collector = DueCreditCollector()
    collector.cite(entry, path="package")
    collector.cite(entry2, path="package2")

    output = Output(None, collector)

    packages, modules, objects = output._get_collated_citations(tags=["*"])
    assert not packages
    assert not modules
    assert not objects

    for flag in ["1", "True", "TRUE", "true", "on", "yes"]:
        monkeypatch.setitem(os.environ, "DUECREDIT_REPORT_ALL", flag)
        # if _all is None then get the environment
        packages, modules, objects = output._get_collated_citations(tags=["*"])
        assert len(packages) == 2
        assert not modules
        assert not objects
        # however if _all is set it shouldn't work
        packages, modules, objects = output._get_collated_citations(
            tags=["*"], all_=False
        )
        assert not packages
        assert not modules
        assert not objects


def test_output_tags(monkeypatch: MonkeyPatch) -> None:
    entry = BibTeX(_sample_bibtex)
    entry2 = BibTeX(_sample_bibtex2)

    # normal use
    collector = DueCreditCollector()
    collector.cite(entry, path="package", cite_module=True, tags=["edu"])
    collector.cite(entry2, path="package.module", tags=["wip"])

    output = Output(None, collector)

    packages, modules, objects = output._get_collated_citations(tags=["*"])
    assert len(packages) == 1
    assert len(modules) == 1
    assert not objects

    packages, modules, objects = output._get_collated_citations()
    assert not packages
    assert not modules
    assert not objects

    for tags in ["edu", "wip", "edu,wip"]:
        monkeypatch.setitem(os.environ, "DUECREDIT_REPORT_TAGS", tags)
        # if tags is None then get the environment
        packages, modules, objects = output._get_collated_citations()
        assert len(packages) == (1 if "edu" in tags else 0)
        assert len(modules) == (1 if "wip" in tags else 0)
        assert not objects
        # however if tags is set it shouldn't work
        packages, modules, objects = output._get_collated_citations(
            tags=["implementation"]
        )
        assert not packages
        assert not modules
        assert not objects


def test_text_output() -> None:
    entry = BibTeX(_sample_bibtex)
    entry2 = BibTeX(_sample_bibtex2)

    # in this case, since we're not citing any module or method, we shouldn't
    # output anything
    collector = DueCreditCollector()
    collector.cite(entry, path="package")

    strio = StringIO()
    TextOutput(strio, collector).dump(tags=["*"])
    value = strio.getvalue()
    assert "0 packages cited" in value, "value was %s" % value
    assert "0 modules cited" in value, "value was %s" % value
    assert "0 functions cited" in value, "value was %s" % value

    # but it should be cited if cite_module=True
    collector = DueCreditCollector()
    collector.cite(entry, path="package", cite_module=True)

    strio = StringIO()
    TextOutput(strio, collector).dump(tags=["*"])
    value = strio.getvalue()
    assert "1 package cited" in value, "value was %s" % value
    assert "0 modules cited" in value, "value was %s" % value
    assert "0 functions cited" in value, "value was %s" % value

    # in this case, we should be citing the package since we are also citing a
    # submodule
    collector = DueCreditCollector()
    collector.cite(entry, path="package")
    collector.cite(entry, path="package.module")

    strio = StringIO()
    TextOutput(strio, collector).dump(tags=["*"])
    value = strio.getvalue()
    assert "1 package cited" in value, "value was %s" % value
    assert "1 module cited" in value, "value was %s" % value
    assert "0 functions cited" in value, "value was %s" % value
    assert "Halchenko, Y.O." in value, "value was %s" % value
    assert value.strip().endswith("Frontiers in Neuroinformatics, 6(22).")

    # in this case, we should be citing the package since we are also citing a
    # submodule
    collector = DueCreditCollector()
    collector.cite(entry, path="package")
    collector.cite(entry2, path="package")
    collector.cite(entry, path="package.module")

    strio = StringIO()
    TextOutput(strio, collector).dump(tags=["*"])
    value = strio.getvalue()
    assert "1 package cited" in value, "value was %s" % value
    assert "1 module cited" in value, "value was %s" % value
    assert "0 functions cited" in value, "value was %s" % value
    assert "Halchenko, Y.O." in value, "value was %s" % value
    assert "[1, 2]" in value, "value was %s" % value
    assert "[3]" not in value, "value was %s" % value


def test_text_output_dump_formatting() -> None:
    due = DueCreditCollector()

    # XXX: atm just to see if it spits out stuff
    @due.dcite(
        BibTeX(_sample_bibtex),
        description="solution to life",
        path="mymodule",
        version="0.0.16",
    )
    def mymodule(arg1: Any, kwarg2: Any = "blah") -> Any:
        """docstring"""
        assert arg1 == "magical"
        assert kwarg2 == 1

        @due.dcite(
            BibTeX(_sample_bibtex2),
            description="solution to life",
            path="mymodule:myfunction",
        )
        def myfunction(arg42):
            pass

        myfunction("argh")
        return "load"

    # check we don't have anything output
    strio = StringIO()
    TextOutput(strio, due).dump(tags=["*"])
    value = strio.getvalue()
    assert "0 modules cited" in value, f"value was {value}"
    assert "0 functions cited" in value, f"value was {value}"

    # now we call it -- check it prints stuff
    strio = StringIO()
    mymodule("magical", kwarg2=1)
    TextOutput(strio, due).dump(tags=["*"])
    value = strio.getvalue()
    assert "1 package cited" in value, f"value was {value}"
    assert "1 function cited" in value, f"value was {value}"
    assert "(v 0.0.16)" in value, f"value was {value}"
    assert len(value.split("\n")) == 16, "value was {}".format(len(value.split("\n")))

    # test we get the reference numbering right
    samples_bibtex = [_generate_sample_bibtex() for x in range(6)]
    # this sucks but at the moment it's the only way to have multiple
    # references for a function

    @due.dcite(
        BibTeX(samples_bibtex[0]),
        description="another solution",
        path="myothermodule",
        version="0.0.666",
    )
    def myothermodule(arg1: Any, kwarg2: Any = "blah") -> Any:
        """docstring"""
        assert arg1 == "magical"
        assert kwarg2 == 1

        @due.dcite(
            BibTeX(samples_bibtex[1]),
            description="solution to life",
            path="myothermodule:myotherfunction",
        )
        @due.dcite(
            BibTeX(samples_bibtex[2]),
            description="solution to life",
            path="myothermodule:myotherfunction",
        )
        @due.dcite(
            BibTeX(samples_bibtex[3]),
            description="solution to life",
            path="myothermodule:myotherfunction",
        )
        @due.dcite(
            BibTeX(samples_bibtex[4]),
            description="solution to life",
            path="myothermodule:myotherfunction",
        )
        @due.dcite(
            BibTeX(samples_bibtex[5]),
            description="solution to life",
            path="myothermodule:myotherfunction",
        )
        def myotherfunction(arg42):
            pass

        myotherfunction("argh")
        return "load"

    myothermodule("magical", kwarg2=1)
    strio = StringIO()
    TextOutput(strio, due).dump(tags=["*"])
    value = strio.getvalue()
    lines = value.split("\n")

    citation_numbers: list[str] = []
    reference_numbers = []
    references = []
    for line in lines:
        match_citation = re.search(r"\[([0-9, ]+)\]$", line)
        match_reference = re.search(r"^\[([0-9])\]", line)
        if match_citation:
            citation_numbers.extend(match_citation.group(1).split(", "))
        elif match_reference:
            reference_numbers.append(match_reference.group(1))
            references.append(line.replace(match_reference.group(), ""))

    assert set(citation_numbers) == set(reference_numbers)
    assert len(set(references)) == len(set(citation_numbers))
    assert len(citation_numbers) == 8
    # verify that we have returned to previous state of filters
    import warnings

    assert ("ignore", None, UserWarning, None, 0) not in warnings.filters


def test_bibtex_output() -> None:
    entry = BibTeX(_sample_bibtex)
    entry2 = BibTeX(_sample_bibtex2)

    # in this case, since we're not citing any module or method, we shouldn't
    # output anything
    collector = DueCreditCollector()
    collector.cite(entry, path="package")

    strio = StringIO()
    BibTeXOutput(strio, collector).dump(tags=["*"])
    value = strio.getvalue()
    assert value == "", f"Value was {value}"

    # impose citing
    collector = DueCreditCollector()
    collector.cite(entry, path="package", cite_module=True)

    strio = StringIO()
    BibTeXOutput(strio, collector).dump(tags=["*"])
    value = strio.getvalue()
    assert value.strip() == _sample_bibtex.strip(), f"Value was {value}"

    # impose filtering
    collector = DueCreditCollector()
    collector.cite(entry, path="package", cite_module=True, tags=["edu"])
    collector.cite(entry2, path="package.module")

    strio = StringIO()
    BibTeXOutput(strio, collector).dump(tags=["edu"])
    value = strio.getvalue()
    assert value.strip() == _sample_bibtex.strip(), f"Value was {value}"

    # no filtering
    strio = StringIO()
    BibTeXOutput(strio, collector).dump(tags=["*"])
    value = strio.getvalue()
    assert (
        value.strip() == _sample_bibtex.strip() + _sample_bibtex2.rstrip()
    ), f"Value was {value}"

    # check the we output only unique bibtex entries
    collector.cite(entry2, path="package")
    strio = StringIO()
    BibTeXOutput(strio, collector).dump(tags=["*"])
    value = strio.getvalue()
    value_ = sorted(value.strip().split("\n"))
    bibtex = sorted((_sample_bibtex.strip() + _sample_bibtex2.rstrip()).split("\n"))
    assert value_ == bibtex, f"Value was {value_} instead of {bibtex}"

    # assert_equal(value_, bibtex,
    #              msg='Value was {0}'.format(value_, bibtex))


def _generate_sample_bibtex() -> str:
    """
    Generate a random sample bibtex to test multiple references
    """
    letters = "abcdefghilmnopqrstuvxz"
    numbers = "0123456789"
    letters_numbers = letters + letters.upper() + numbers
    letters_numbers_spaces = letters_numbers + " "

    key = "".join(random.sample(letters_numbers, 7))
    title = "".join(random.sample(letters_numbers_spaces, 20))
    journal = "".join(random.sample(letters_numbers_spaces, 20))
    publisher = "".join(random.sample(letters_numbers_spaces, 10))
    author = (
        "".join(random.sample(letters, 6)) + ", " + "".join(random.sample(letters, 4))
    )
    year = "".join(random.sample(numbers, 4))

    elements = [
        ("title", title),
        ("journal", journal),
        ("publisher", publisher),
        ("author", author),
        ("year", year),
    ]

    sample_bibtex = "@ARTICLE{%s,\n" % key
    for string, value in elements:
        sample_bibtex += "{}={{{}}},\n".format(string, value)
    sample_bibtex += "}"
    return sample_bibtex


def test_get_text_rendering(monkeypatch: MonkeyPatch) -> None:
    # Patch bibtex_rendering
    sample_bibtex = BibTeX(_sample_bibtex)

    def get_bibtex_rendering(*_args: Any, **_kwargs: Any) -> BibTeX:
        return sample_bibtex

    monkeypatch.setattr(duecredit.io, "get_bibtex_rendering", get_bibtex_rendering)

    # Patch format_bibtex
    fmt_args = {}

    def format_bibtex(entry: str, style: str) -> None:
        fmt_args["entry"] = entry
        fmt_args["style"] = style

    monkeypatch.setattr(duecredit.io, "format_bibtex", format_bibtex)

    # test if bibtex type is passed
    bibtex_output = get_text_rendering(sample_bibtex)
    assert fmt_args["entry"] == sample_bibtex
    assert fmt_args["style"] == "harvard1"
    fmt_args.clear()

    # test if doi type is passed
    doi_output = get_text_rendering(Doi(_sample_doi))
    assert fmt_args["entry"] == sample_bibtex
    assert fmt_args["style"] == "harvard1"

    assert bibtex_output == doi_output


def test_text_text_rendering() -> None:
    text = "I am so free"
    assert get_text_rendering(Text(text)) == text


def test_url_text_rendering() -> None:
    url = "http://example.com"
    assert get_text_rendering(Url(url)) == "URL: " + url


def test_format_bibtex_zenodo_doi() -> None:
    """
    test that we can correctly parse bibtex entries obtained from a zenodo doi
    """
    # this was fetched on 2016-05-10
    bibtex_zenodo = """
    @data{0b1284ba-5ce5-4367-84f3-c44b4962ad90,
    doi = {10.5281/zenodo.50186},
    url = {https://doi.org/10.5281/zenodo.50186},
    author = {Satrajit Ghosh; Chris Filo Gorgolewski; Oscar Esteban;
    Erik Ziegler; David Ellis; cindeem; Michael Waskom; Dav Clark; Michael;
    Fred Loney; Alexandre M. S.; Michael Notter; Hans Johnson;
    Anisha Keshavan; Yaroslav Halchenko; Carlo Hamalainen; Blake Dewey;
    Ben Cipollini; Daniel Clark; Julia Huntenburg; Drew Erickson;
    Michael Hanke; moloney; Jason W; Demian Wassermann; cdla;
    Nolan Nichols; Chris Markiewicz; Jarrod Millman; Arman Eshaghi; },
    publisher = {Zenodo},
    title = {nipype: Release candidate 1 for version 0.12.0},
    year = {2016}
    }
    """
    assert (
        format_bibtex(BibTeX(bibtex_zenodo))
        == """Ghosh, S. et al., 2016. nipype: Release candidate 1 for version 0.12.0."""
    )


def test_format_bibtex_with_utf_characters() -> None:
    """
    test that we can correctly parse bibtex entry if it contains utf-8 characters
    """
    # this was fetched on 2017-08-16
    # replaced Brett with Brótt to have utf-8 characters in first author's name as well
    bibtex_utf8 = (
        "@misc{https://doi.org/10.5281/zenodo.60847,\n  doi = {10.5281/zenodo.60847},\n  url = {"
        "http://zenodo.org/record/60847},\n  author = {Brótt, Matthew and Hanke, Michael and Cipollini, "
        "Ben and {Marc-Alexandre Côté} and Markiewicz, Chris and Gerhard, Stephan and Larson, "
        "Eric and Lee, Gregory R. and Halchenko, Yaroslav and Kastman, Erik and {Cindeem} and Morency, "
        "Félix C. and {Moloney} and Millman, Jarrod and Rokem, Ariel and {Jaeilepp} and Gramfort, "
        "Alexandre and Bosch, Jasper J.F. Van Den and {Krish Subramaniam} and Nichols, Nolan and {Embaker} "
        "and {Bpinsard} and {Chaselgrove} and Oosterhof, Nikolaas N. and St-Jean, Samuel and {Bago "
        "Amirbekian} and Nimmo-Smith, Ian and {Satrajit Ghosh}},\n  keywords = {},\n  title = {nibabel "
        "2.0.1},\n  publisher = {Zenodo},\n  year = {2015}\n} "
    )
    assert (
        format_bibtex(BibTeX(bibtex_utf8)) == "Brótt, M. et al., 2015. nibabel 2.0.1."
    )


def test_is_contained() -> None:
    toppath = "package"
    assert _is_contained(toppath, "package.module")
    assert _is_contained(toppath, "package.module.submodule")
    assert _is_contained(toppath, "package.module.submodule:object")
    assert _is_contained(toppath, "package:object")
    assert _is_contained(toppath, toppath)
    assert not _is_contained(toppath, "package2")
    assert not _is_contained(toppath, "package2:anotherobject")
    assert not _is_contained(toppath, "package2.module:anotherobject")



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.  This file originates from datalad distributed
#   under MIT license.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
from __future__ import annotations

import logging
import logging.handlers
import os
from os.path import basename, dirname
import platform
import re
import sys
import traceback

from .utils import is_interactive

__all__ = ["ColorFormatter", "lgr"]

# Snippets from traceback borrowed from PyMVPA upstream/2.4.0-39-g69ad545  MIT license


def mbasename(s: str) -> str:
    """Custom function to include directory name if filename is too common

    Also strip .py at the end
    """
    base = basename(s)
    if base.endswith(".py"):
        base = base[:-3]
    if base in {"base", "__init__"}:
        base = basename(dirname(s)) + "." + base
    return base


class TraceBack:
    """Customized traceback to be included in debug messages"""

    def __init__(self, collide: bool = False) -> None:
        """Initialize TraceBack metric

        Parameters
        ----------
        collide : bool
          if True then prefix common with previous invocation gets
          replaced with ...
        """
        self.__prev = ""
        self.__collide = collide

    def __call__(self) -> str:
        ftb = traceback.extract_stack(limit=100)[:-2]
        entries = [
            [mbasename(x[0]), str(x[1])]
            for x in ftb
            if mbasename(x[0]) != "logging.__init__"
        ]
        entries = [e for e in entries if e[0] != "unittest"]

        # lets make it more concise
        entries_out = [entries[0]]
        for entry in entries[1:]:
            if entry[0] == entries_out[-1][0]:
                entries_out[-1][1] += ",%s" % entry[1]
            else:
                entries_out.append(entry)
        sftb = ">".join(["{}:{}".format(mbasename(x[0]), x[1]) for x in entries_out])
        if self.__collide:
            # lets remove part which is common with previous invocation
            prev_next = sftb
            common_prefix = os.path.commonprefix((self.__prev, sftb))
            common_prefix2 = re.sub(">[^>]*$", "", common_prefix)

            if common_prefix2 != "":
                sftb = "..." + sftb[len(common_prefix2) :]
            self.__prev = prev_next

        return sftb


# Recipe from http://stackoverflow.com/questions/384076/how-can-i-color-python-logging-output
# by Brandon Thomson
# Adjusted for automagic determination either coloring is needed and
# prefixing of multiline log lines
class ColorFormatter(logging.Formatter):
    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

    RESET_SEQ = "\033[0m"
    COLOR_SEQ = "\033[1;%dm"
    BOLD_SEQ = "\033[1m"

    COLORS = {
        "WARNING": YELLOW,
        "INFO": WHITE,
        "DEBUG": BLUE,
        "CRITICAL": YELLOW,
        "ERROR": RED,
    }

    def __init__(self, use_color: bool | None = None, log_name: bool = False) -> None:
        if use_color is None:
            # if 'auto' - use color only if all streams are tty
            use_color = is_interactive()
        self.use_color = (
            use_color and platform.system() != "Windows"
        )  # don't use color on windows
        msg = self.formatter_msg(self._get_format(log_name), self.use_color)
        self._tb = (
            TraceBack(collide=os.environ.get("DUECREDIT_LOGTRACEBACK", "") == "collide")
            if os.environ.get("DUECREDIT_LOGTRACEBACK", False)
            else None
        )
        logging.Formatter.__init__(self, msg)

    def _get_format(self, log_name: bool = False) -> str:
        return (
            "$BOLD%(asctime)-15s$RESET "
            + ("%(name)-15s " if log_name else "")
            + "[%(levelname)s] "
            "%(message)s "
            "($BOLD%(filename)s$RESET:%(lineno)d)"
        )

    def formatter_msg(self, fmt: str, use_color: bool = False) -> str:
        if use_color:
            fmt = fmt.replace("$RESET", self.RESET_SEQ).replace("$BOLD", self.BOLD_SEQ)
        else:
            fmt = fmt.replace("$RESET", "").replace("$BOLD", "")
        return fmt

    def format(self, record: logging.LogRecord) -> str:
        if record.msg.startswith("| "):
            # If we already log smth which supposed to go without formatting, like
            # output for running a command, just return the message and be done
            return record.msg

        levelname = record.levelname
        if self.use_color and levelname in self.COLORS:
            fore_color = 30 + self.COLORS[levelname]
            levelname_color = (
                self.COLOR_SEQ % fore_color + "%-7s" % levelname + self.RESET_SEQ
            )
            record.levelname = levelname_color
        record.msg = record.msg.replace("\n", "\n| ")
        if self._tb:
            record.msg = self._tb() + "  " + record.msg

        return logging.Formatter.format(self, record)


class LoggerHelper:
    """Helper to establish and control a Logger"""

    def __init__(self, name: str = "duecredit") -> None:
        self.name = name
        self.lgr = logging.getLogger(name)

    def _get_environ(
        self, var: str, default: str | None | bool = None
    ) -> str | None | bool:
        return os.environ.get(self.name.upper() + "_%s" % var.upper(), default)

    def set_level(self, level: str | None = None, default: str = "WARNING") -> None:
        """Helper to set loglevel for an arbitrary logger

        By default operates for 'duecredit'.
        TODO: deduce name from upper module name so it could be reused without changes
        """
        if level is None:
            # see if nothing in the environment
            lv = self._get_environ("LOGLEVEL")
            assert type(lv) is not bool
            level = lv
        if level is None:
            level = default

        try:
            # it might be a string which still represents an int
            log_level = int(level)
        except ValueError:
            # or a string which corresponds to a constant;)
            log_level = getattr(logging, level.upper())

        self.lgr.setLevel(log_level)

    def get_initialized_logger(self, logtarget: str | None = None) -> logging.Logger:
        """Initialize and return the logger

        Parameters
        ----------
        target: string, optional
          Which log target to request logger for
        logtarget: { 'stderr', 'stdout', str }, optional
          Where to direct the logs.  stdout and stderr stand for standard streams.
          Any other string is considered a filename.  Multiple entries could be
          specified comma-separated

        Returns
        -------
        logging.Logger
        """
        # By default mimic previously talkative behavior
        logtarget_ = self._get_environ("LOGTARGET", logtarget or "stderr")
        assert type(logtarget_) is str

        # Allow for multiple handlers being specified, comma-separated
        if "," in logtarget_:
            for handler_ in logtarget_.split(","):
                self.get_initialized_logger(logtarget=handler_)
            return self.lgr

        if logtarget_.lower() in ("stdout", "stderr"):
            loghandler = logging.StreamHandler(getattr(sys, logtarget_.lower()))
            use_color = is_interactive()  # explicitly decide here
        else:
            # must be a simple filename
            # Use RotatingFileHandler for possible future parametrization to keep
            # log succinct and rotating
            loghandler = logging.handlers.RotatingFileHandler(logtarget_)
            use_color = False
            # I had decided not to guard this call and just raise exception to go
            # out happen that specified file location is not writable etc.
        # But now improve with colors and useful information such as time
        log_name = self._get_environ("LOGNAME", False)
        assert type(log_name) is bool
        loghandler.setFormatter(ColorFormatter(use_color=use_color, log_name=log_name))
        # logging.Formatter('%(asctime)-15s %(levelname)-6s %(message)s'))
        self.lgr.addHandler(loghandler)

        self.set_level()  # set default logging level
        return self.lgr


lgr = LoggerHelper().get_initialized_logger()



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
from __future__ import annotations

from collections.abc import Iterator
import os
from os.path import dirname
from os.path import join as pathjoin
import shutil
from subprocess import PIPE, Popen
import sys
import tempfile
from typing import overload

import pytest
from pytest import MonkeyPatch

from duecredit.collector import DueCreditCollector
from duecredit.entries import BibTeX, Doi
from duecredit.stub import InactiveDueCreditCollector

from ..utils import on_windows

# temporary location where stuff would be copied
badlxml_path = pathjoin(dirname(__file__), "envs", "nolxml")
stubbed_dir = tempfile.mktemp()
stubbed_script = pathjoin(pathjoin(stubbed_dir, "script.py"))


@pytest.fixture(scope="module")
def stubbed_env() -> Iterator[str]:
    """Create stubbed module with a sample script"""
    os.makedirs(stubbed_dir)
    with open(stubbed_script, "wb") as f:
        f.write(
            b"""
from due import due, Doi

kwargs = dict(
    entry=Doi("10.1007/s12021-008-9041-y"),
    description="Multivariate pattern analysis of neural data",
    tags=["use"]
)

due.cite(path="test", **kwargs)


@due.dcite(**kwargs)
def method(arg):
    return arg+1

assert method(1) == 2
print("done123")
"""
        )
    # copy stub.py under stubbed
    shutil.copy(
        pathjoin(dirname(__file__), os.pardir, "stub.py"),
        pathjoin(stubbed_dir, "due.py"),
    )
    yield stubbed_script
    # cleanup
    shutil.rmtree(stubbed_dir)


@pytest.mark.parametrize(
    "collector_class", [DueCreditCollector, InactiveDueCreditCollector]
)
def test_api(collector_class) -> None:
    due = collector_class()
    # add references
    due.add(BibTeX("@article{XXX00, ...}"))
    # could even be by DOI -- we need to fetch and cache those
    due.add(Doi("xxx.yyy/zzz.1", key="XXX01"))

    # and/or load multiple from a file
    due.load("/home/siiioul/deep/good_intentions.bib")

    # Cite entire module
    due.cite("XXX00", description="Answers to existential questions", path="module")
    # Cite some method within some submodule
    due.cite(
        "XXX01",
        description="More answers to existential questions",
        path="module.submodule:class1.whoknowswhat2.func1",
    )

    # dcite  for decorator cite
    # cite specific functionality if/when it gets called up
    @due.dcite("XXX00", description="Provides an answer for meaningless existence")
    def purpose_of_life():
        return None

    class Child:
        # Conception process is usually way too easy to be referenced
        def __init__(self) -> None:
            pass

        # including functionality within/by the methods
        @due.dcite("XXX00")
        def birth(self, _gender) -> str:
            return "Rachel was born"

    kid = Child()
    kid.birth("female")


@overload
def run_python_command(cmd: str):
    ...


@overload
def run_python_command(cmd: None, script: str):
    ...


def run_python_command(cmd: str | None = None, script: str | None = None):
    """Just a tiny helper which runs command and returns exit code, stdout, stderr"""
    if script is None:
        assert cmd is not None
        args = ["-c", cmd]
    else:
        assert cmd is None
        args = [script]
    try:
        # run script from some temporary directory so we do not breed .duecredit.p
        # in current directory
        tmpdir = tempfile.mkdtemp()
        python = Popen([sys.executable] + args, stdout=PIPE, stderr=PIPE, cwd=tmpdir)
        stdout, stderr = python.communicate()  # wait()
        ret = python.poll()
    finally:
        shutil.rmtree(tmpdir)
    # TODO stdout cannot decode on Windows special character /x introduced
    return ret, stdout.decode(errors="ignore"), stderr.decode()


# Since duecredit and possibly lxml already loaded, let's just test
# ability to import in absence of lxml via external call to python
def test_noincorrect_import_if_no_lxml(monkeypatch: MonkeyPatch) -> None:
    if on_windows:
        pytest.xfail("Fails for some reason on Windows")

    monkeypatch.setitem(
        os.environ,
        "PYTHONPATH",
        "{}:{}".format(badlxml_path, os.environ.get("PYTHONPATH", "")),
    )
    ret, out, err = run_python_command("import lxml")
    assert ret == 1
    assert "ImportError" in err

    ret, out, err = run_python_command("import duecredit")
    if "CoverageWarning" not in err:
        # TODO: deal with that warning
        # "--include is ignored because --source is set"
        assert err == ""
    assert out == ""
    assert ret == 0


@pytest.mark.parametrize(
    "env",
    [
        {},
        {"DUECREDIT_ENABLE": "yes"},
        {"DUECREDIT_ENABLE": "yes", "DUECREDIT_REPORT_TAGS": "*"},
        {"DUECREDIT_TEST_EARLY_IMPORT_ERROR": "yes"},
    ],
)
@pytest.mark.parametrize(
    "kwargs",
    [
        # direct command to evaluate
        {"cmd": 'import duecredit; import numpy as np; print("done123")'},
        # script with decorated funcs etc -- should be importable
        {"script": stubbed_script},
    ],
)
def test_noincorrect_import_if_no_lxml_numpy(
    monkeypatch: MonkeyPatch, kwargs, env, stubbed_env  # noqa: U100
) -> None:
    # Now make sure that we would not crash entire process at the end when unable to
    # produce sensible output when we have something to cite
    # we do inject for numpy
    try:
        import numpy  # noqa: F401
    except ImportError:
        pytest.skip("We need to have numpy to test correct operation")

    fake_env_nolxml_ = {
        "PYTHONPATH": "{}:{}".format(badlxml_path, os.environ.get("PYTHONPATH", ""))
    }.copy()
    fake_env_nolxml_.update(env)

    for key in fake_env_nolxml_:
        monkeypatch.setitem(os.environ, key, fake_env_nolxml_[key])

    ret, out, err = run_python_command(**kwargs)
    direct_duecredit_import = "import duecredit" in kwargs.get("cmd", "")
    if direct_duecredit_import and env.get("DUECREDIT_TEST_EARLY_IMPORT_ERROR", ""):
        # We do fail then upon regular import but stubbed script should be ok
        # since should use the stub
        assert "Both inactive and active collectors should be provided" in err
        assert ret == 1
    else:
        assert err == ""
        assert ret == 0  # but we must not fail overall regardless

    if (
        os.environ.get("DUECREDIT_ENABLE", False) and on_windows
    ):  # TODO this test fails on windows
        pytest.xfail("Fails for some reason on Windows")
    elif os.environ.get("DUECREDIT_ENABLE", False):  # we enabled duecredit
        if (
            os.environ.get("DUECREDIT_REPORT_TAGS", None) == "*"
            and kwargs.get("script")
        ) or "numpy" in kwargs.get("cmd", ""):
            # we requested to have all tags output, and used bibtex in our entry
            assert "For formatted output we need citeproc" in out
        else:
            # there was nothing to format so we did not fail for no reason
            assert "For formatted output we need citeproc" not in out
            assert "0 packages cited" in out
        assert "done123" in out
    elif os.environ.get("DUECREDIT_TEST_EARLY_IMPORT_ERROR"):
        assert "ImportError" in out
        assert "DUECREDIT_TEST_EARLY_IMPORT_ERROR" in out
        if direct_duecredit_import:
            assert "Please report" in out
        else:
            assert "done123" in out
    else:
        assert "done123\n" or "done123\r\n" == out


if __name__ == "__main__":
    from duecredit import due

    test_api(due)



# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from ..entries import Doi, Text, Url


def test_comparison() -> None:
    assert Text("123") == Text("123")
    assert Text("123") != Text("124")
    assert Text("123", "key") == Text("123", "key")
    assert Text("123", "key") != Text("123", "key1")
    assert Text("123", "key") != Text("124", "key")

    assert Doi("123/1", "key") == Doi("123/1", "key")
    assert Url("http://123/1", "key") == Url("http://123/1", "key")


def test_sugaring_api() -> None:
    assert Url("http://1.com").url == "http://1.com"
    assert Doi("1.com").doi == "1.com"



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
"""
from __future__ import annotations

__docformat__ = "restructuredtext"

import argparse
import re
import sys
from typing import Any, Pattern

from ..utils import is_interactive


class HelpAction(argparse.Action):
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,  # noqa: U100
        values,  # noqa: U100
        option_string: str | None = None,
    ) -> None:
        # import pydb; pydb.debugger()

        if is_interactive() and option_string == "--help":
            # lets use the manpage on mature systems ...
            try:
                import subprocess

                subprocess.check_call(
                    "man %s 2> /dev/null" % parser.prog.replace(" ", "-"),
                    shell=True,
                )
                sys.exit(0)
            except (subprocess.CalledProcessError, OSError):
                # ...but silently fall back if it doesn't work
                pass
        if option_string == "-h":
            helpstr = "%s\n%s" % (
                parser.format_usage(),
                "Use '--help' to get more comprehensive information.",
            )
        else:
            helpstr = parser.format_help()
        # better for help2man
        helpstr = re.sub(r"optional arguments:", "options:", helpstr)
        # yoh: TODO for duecredit + help2man
        # helpstr = re.sub(r'positional arguments:\n.*\n', '', helpstr)
        # convert all heading to have the first character uppercase
        headpat = re.compile(r"^([a-z])(.*):$", re.MULTILINE)
        helpstr = re.sub(
            headpat, lambda match: rf"{match[1].upper()}{match[2]}:", helpstr
        )
        # usage is on the same line
        helpstr = re.sub(r"^usage:", "Usage:", helpstr)
        if option_string == "--help-np":
            usagestr = re.split(r"\n\n[A-Z]+", helpstr, maxsplit=1)[0]
            usage_length = len(usagestr)
            usagestr = re.subn(r"\s+", " ", usagestr.replace("\n", " "))[0]
            helpstr = "{}\n{}".format(usagestr, helpstr[usage_length:])
        print(helpstr)
        sys.exit(0)


class LogLevelAction(argparse.Action):
    def __call__(
        self,
        parser: argparse.ArgumentParser,  # noqa: U100
        namespace: argparse.Namespace,  # noqa: U100
        values,
        option_string: str | None = None,  # noqa: U100
    ) -> None:
        from ..log import LoggerHelper

        LoggerHelper().set_level(level=values)


def parser_add_common_args(
    parser: argparse.ArgumentParser, pos=None, opt=None, **kwargs: Any
) -> None:
    from . import common_args

    for i, args in enumerate((pos, opt)):
        if args is None:
            continue
        for arg in args:
            arg_tmpl = getattr(common_args, arg)
            arg_kwargs = arg_tmpl[2].copy()
            arg_kwargs.update(kwargs)
            if i:
                parser.add_argument(*arg_tmpl[i], **arg_kwargs)
            else:
                parser.add_argument(arg_tmpl[i], **arg_kwargs)


def parser_add_common_opt(
    parser: argparse.ArgumentParser, opt, names=None, **kwargs
) -> None:
    from . import common_args

    opt_tmpl = getattr(common_args, opt)
    opt_kwargs = opt_tmpl[2].copy()
    opt_kwargs.update(kwargs)
    if names is None:
        parser.add_argument(*opt_tmpl[1], **opt_kwargs)
    else:
        parser.add_argument(*names, **opt_kwargs)


class RegexpType:
    """Factory for creating regular expression types for argparse

    DEPRECATED AFAIK -- now things are in the config file...
    but we might provide a mode where we operate solely from cmdline
    """

    def __call__(self, string: str | None) -> Pattern[str] | None:
        if string:
            return re.compile(string)
        else:
            return None



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.   Originates from datalad package distributed
#   under MIT license
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
from __future__ import annotations

import sys
from typing import Any

from pytest import MonkeyPatch

from ..utils import is_interactive


def test_is_interactive_crippled_stdout(monkeypatch: MonkeyPatch) -> None:
    class MockedOut:
        """the one which has no isatty"""

        def write(self, *args: Any, **kwargs: Any) -> None:
            pass

    class MockedIsaTTY(MockedOut):
        def isatty(self) -> bool:
            return True

    for inout in ("in", "out", "err"):
        monkeypatch.setattr(sys, "std%s" % inout, MockedOut())
        assert not is_interactive()

    # just for paranoids
    for inout in ("in", "out", "err"):
        monkeypatch.setattr(sys, "std%s" % inout, MockedIsaTTY())
    assert is_interactive()



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
from __future__ import annotations

import atexit
from typing import Any

import pytest
from pytest import MonkeyPatch

from ..dueswitch import DueSwitch, due
from ..injections.injector import DueCreditInjector


def test_dueswitch_activate(monkeypatch: MonkeyPatch) -> None:
    if due.active:
        pytest.skip("due is already active, can't test more at this point")

    # TODO: use TypedDict PEP 692 or a small typed object
    state = dict(activate=0, register=0, register_func=None)

    # Patch DueCreditInjector.activate
    def activate_calls(*_args: Any, **_kwargs: Any) -> None:
        assert type(state["activate"]) is int
        state["activate"] += 1

    monkeypatch.setattr(DueCreditInjector, "activate", activate_calls)

    # Patch atexit.register
    def register(func) -> None:
        assert type(state["register"]) is int
        state["register"] += 1
        state["register_func"] = func

    monkeypatch.setattr(atexit, "register", register)

    due.activate()

    # was not active, so should have called activate of the injector class
    assert state["activate"] == 1
    assert state["register"] == 1
    assert state["register_func"] == due.dump


def test_a_bad_one() -> None:
    # We might get neither of those and should fail
    # see https://github.com/duecredit/duecredit/issues/142
    # So let's through ValueError right away
    pytest.raises(ValueError, DueSwitch, None, None, True)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Automatic injection of bibliography entries for pandas module
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..entries import BibTeX

# If defined, would determine from which to which version of the corresponding
# module to care about
min_version = None
max_version = None

if TYPE_CHECKING:
    from .injector import DueCreditInjector


def inject(injector: DueCreditInjector) -> None:
    injector.add(
        "pandas",
        None,
        BibTeX(
            """
        @InProceedings{ mckinney-proc-scipy-2010,
          author    = { McKinney, Wes },
          title     = { Data Structures for Statistical Computing in Python },
          booktitle = { Proceedings of the 9th Python in Science Conference },
          pages     = { 51 -- 56 },
          year      = { 2010 },
          editor    = { van der Walt, St\'efan and Millman, Jarrod }
        }
    """
        ),
        description="Data analysis library for tabular data",
    )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
from __future__ import annotations

from io import StringIO
import sys
from typing import TYPE_CHECKING

import pytest
from pytest import MonkeyPatch

from .. import __main__, __version__, due

if TYPE_CHECKING:
    import py


def test_main_help(monkeypatch: MonkeyPatch) -> None:
    # Patch stdout
    fakestdout = StringIO()
    monkeypatch.setattr(sys, "stdout", fakestdout)

    pytest.raises(SystemExit, __main__.main, ["__main__.py", "--help"])
    assert fakestdout.getvalue().startswith(
        "Usage: %s -m duecredit [OPTIONS] <file> [ARGS]\n" % sys.executable
    )


def test_main_version(monkeypatch: MonkeyPatch) -> None:
    # Patch stdout
    fakestdout = StringIO()
    monkeypatch.setattr(sys, "stdout", fakestdout)

    pytest.raises(SystemExit, __main__.main, ["__main__.py", "--version"])
    assert fakestdout.getvalue().rstrip() == "duecredit %s" % __version__


def test_main_run_a_script(tmpdir: py.path.local, monkeypatch: MonkeyPatch) -> None:
    tempfile = str(tmpdir.mkdir("sub").join("tempfile.txt"))
    content = b'print("Running the script")\n'
    with open(tempfile, "wb") as f:
        f.write(content)

    # Patch stdout
    fakestdout = StringIO()
    monkeypatch.setattr(sys, "stdout", fakestdout)

    # Patch due.activate
    count = [0]

    def count_calls(*_args, **_kwargs):
        count[0] += 1

    monkeypatch.setattr(due, "activate", count_calls)

    __main__.main(["__main__.py", tempfile])
    assert fakestdout.getvalue().rstrip() == "Running the script"

    # And we have "activated" the due
    assert count[0] == 1



from typing import Any


def testfunc1(arg1: Any, kwarg1: Any = None) -> str:
    """custom docstring"""
    return "testfunc1: {}, {}".format(arg1, kwarg1)


class TestClass1:
    """wrong custom docstring"""

    def testmeth1(self, arg1: Any, kwarg1: Any = None) -> str:
        """custom docstring"""
        return "TestClass1.testmeth1: {}, {}".format(arg1, kwarg1)


class TestClass12:
    """wrong custom docstring"""

    class Embed:
        """wrong custom docstring"""

        def testmeth1(self, arg1: Any, kwarg1: Any = None) -> str:
            """custom docstring"""
            return "TestClass12.Embed.testmeth1: {}, {}".format(arg1, kwarg1)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
Automatic injection of bibliography entries for matplotlib module
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..entries import BibTeX, Doi

# If defined, would determine from which to which version of the corresponding
# module to care about
min_version = None
max_version = None

if TYPE_CHECKING:
    from .injector import DueCreditInjector

bib_str = """
@Article{Hunter:2007,
  Author    = {Hunter, J. D.},
  Title     = {Matplotlib: A 2D graphics environment},
  Journal   = {Computing in Science \\& Engineering},
  Volume    = {9},
  Number    = {3},
  Pages     = {90--95},
  abstract  = {Matplotlib is a 2D graphics package used for Python for
  application development, interactive scripting, and publication-quality
  image generation across user interfaces and operating systems.},
  publisher = {IEEE COMPUTER SOC},
  doi       = {10.1109/MCSE.2007.55},
  year      = 2007
}
""".strip()


def inject(injector: DueCreditInjector) -> None:
    injector.add(
        "matplotlib",
        None,
        BibTeX(bib_str),
        description="Plotting with Python",
        tags=["implementation"],
    )

    doi_prefix = "10.5281/zenodo."

    # latest version
    injector.add(
        "matplotlib",
        None,
        Doi(doi_prefix + "2893252"),
        description="Plotting with Python",
        tags=["implementation"],
    )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Automatic injection of bibliography entries for biosig module
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..entries import Doi

# If defined, would determine from which to which version of the corresponding
# module to care about
min_version = None
max_version = None

if TYPE_CHECKING:
    from .injector import DueCreditInjector


def inject(injector: DueCreditInjector) -> None:
    injector.add(
        "biosig",
        None,
        Doi("10.1109/MC.2008.407"),
        description="I/O library for biosignal data formats",
    )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any

import pytest

from ..collector import Citation, CitationKey, CollectorSummary, DueCreditCollector
from ..dueswitch import DueSwitch
from ..entries import BibTeX, Doi
from ..io import PickleOutput
from ..stub import InactiveDueCreditCollector

if TYPE_CHECKING:
    import py


def _test_entry(due, entry) -> None:
    due.add(entry)


_sample_bibtex = """
@ARTICLE{XXX0,
  author = {Halchenko, Yaroslav O. and Hanke, Michael},
  title = {Open is not enough. Let{'}s take the next step: An integrated, community-driven
    computing platform for neuroscience},
  journal = {Frontiers in Neuroinformatics},
  year = {2012},
  volume = {6},
  number = {00022},
  doi = {10.3389/fninf.2012.00022},
  issn = {1662-5196},
  localfile = {HH12.pdf},
}
"""
_sample_bibtex2 = """
@ARTICLE{Atkins_2002,
  title = {title},
  volume = {666},
  url = {https://doi.org/10.1038/nrd842},
  DOI = {10.1038/nrd842},
  number = {3009},
  journal = {My Fancy. Journ.},
  publisher = {The Publisher},
  author = {Atkins, Joshua H. and Gershell, Leland J.},
  year = {2002},
  month = {Jul},
}
"""
_sample_doi = "10.3389/fninf.2012.00022"


def test_citation_paths() -> None:
    entry = BibTeX(_sample_bibtex)

    cit1 = Citation(entry, path="somemodule")
    assert cit1.cites_module
    assert cit1.module == "somemodule"

    cit2 = Citation(entry, path="somemodule.submodule")
    assert cit2.cites_module
    assert cit2.module == "somemodule.submodule"

    assert cit1 in cit1
    assert cit2 in cit1
    assert cit1 not in cit2

    cit3 = Citation(entry, path="somemodule.submodule:class2.func2")
    assert not cit3.cites_module
    assert cit3.module == "somemodule.submodule"

    assert cit2 in cit1
    assert cit3 in cit1
    assert cit3 in cit2
    assert cit2 not in cit3

    cit4 = Citation(entry, path="somemodule2:class2.func2")
    assert not cit4.cites_module
    assert cit4.module == "somemodule2"

    assert cit1 not in cit4
    assert cit4 not in cit1


def test_entry() -> None:
    entry = BibTeX(_sample_bibtex)
    _test_entry(DueCreditCollector(), entry)

    entries = [BibTeX(_sample_bibtex), BibTeX(_sample_bibtex), Doi(_sample_doi)]
    _test_entry(DueCreditCollector(), entries)


def _test_dcite_basic(_due, func) -> None:
    assert func("magical", 1) == "load"
    # verify that @wraps correctly passes all the docstrings etc
    assert func.__name__ == "method"
    assert func.__doc__ == "docstring"


def test_dcite_method() -> None:
    # Test basic wrapping that we don't mask out the arguments
    for due in [DueCreditCollector(), InactiveDueCreditCollector()]:
        assert isinstance(due, (DueCreditCollector, InactiveDueCreditCollector))
        active = isinstance(due, DueCreditCollector)
        due.add(BibTeX(_sample_bibtex))

        @due.dcite("XXX0", path="method")
        def method(arg1: str, kwarg2: str = "blah") -> str:
            """docstring"""
            assert arg1 == "magical"
            assert kwarg2 == 1
            return "load"

        class SomeClass:
            @due.dcite("XXX0", path="someclass:method")  # type: ignore
            def method(self, arg1: str, kwarg2: str = "blah") -> str:
                """docstring"""
                assert arg1 == "magical"
                assert kwarg2 == 1
                return "load"

        if active:
            assert isinstance(due, DueCreditCollector)
            assert due.citations == {}
            assert len(due._entries) == 1

        _test_dcite_basic(due, method)

        if active:
            assert isinstance(due, DueCreditCollector)
            assert len(due.citations) == 1
            assert len(due._entries) == 1
            citation = due.citations[CitationKey("method", "XXX0")]
            assert citation.count == 1
            # TODO: this is probably incomplete path but unlikely we would know
            # any better
            assert citation.path == "method"

        instance = SomeClass()
        _test_dcite_basic(due, instance.method)

        if active:
            assert isinstance(due, DueCreditCollector)
            assert len(due.citations) == 2
            assert len(due._entries) == 1
            # TODO: we should actually get path/counts pairs so here
            citation = due.citations[CitationKey("someclass:method", "XXX0")]
            assert citation.path == "someclass:method"
            assert citation.count == 1

            # And we explicitly stated that module need to be cited
            assert citation.cite_module

        class SomeClass2:
            # Used to test for classes that are not instantiated
            @due.dcite("XXX0", path="some.module.without.method")  # type: ignore
            def method2(self, arg1, kwarg2="blah"):  # noqa: U100
                assert arg1 == "magical"
                return "load"

        # and a method pointing to the module
        instance2 = SomeClass()

        _test_dcite_basic(due, instance2.method)
        if active:
            assert isinstance(due, DueCreditCollector)
            assert len(due.citations) == 2  # different paths
            assert len(due._entries) == 1  # the same entry
            # TODO: we should actually get path/counts pairs so here
            # it is already a different path
            # And we still explicitly stated that module need to be cited
            assert citation.cite_module


def _test_args_match_conditions(conds: dict[tuple[int, str], set[str]]) -> None:
    args_match_conditions = DueCreditCollector._args_match_conditions
    assert args_match_conditions(conds)
    assert args_match_conditions(conds, None)
    assert args_match_conditions(conds, someirrelevant=True)
    assert args_match_conditions(conds, method="purge")
    assert args_match_conditions(conds, method="fullpurge")
    assert args_match_conditions(conds, None, "purge")
    assert args_match_conditions(conds, None, "fullpurge")
    assert args_match_conditions(conds, None, "fullpurge", someirrelevant="buga")
    assert not args_match_conditions(conds, None, "push")
    assert not args_match_conditions(conds, method="push")
    if len(conds) < 2:
        return
    #  got compound case
    assert args_match_conditions(conds, scope="life")
    assert not args_match_conditions(conds, scope="someother")
    # should be "and", so if one not matching -- both not matching
    assert not args_match_conditions(conds, method="wrong", scope="life")
    assert not args_match_conditions(conds, method="purge", scope="someother")
    # assert args_match_conditions(conds, None, None, 'life')  # ambiguous/conflicting


def test_args_match_conditions() -> None:
    _test_args_match_conditions({(1, "method"): {"purge", "fullpurge", "DC_DEFAULT"}})
    _test_args_match_conditions(
        {
            (1, "method"): {"purge", "fullpurge", "DC_DEFAULT"},
            (2, "scope"): {"life", "DC_DEFAULT"},
        }
    )


def _test_dcite_match_conditions(due, func, path: str) -> None:
    assert due.citations == {}
    assert len(due._entries) == 1

    assert func("magical", "unknown") == "load unknown"
    assert due.citations == {}
    assert len(due._entries) == 1

    assert func("magical") == "load blah"

    assert len(due.citations) == 1
    assert len(due._entries) == 1
    entry = due._entries["XXX0"]  # noqa: F841
    assert due.citations[(path, "XXX0")].count == 1

    # Cause the same citation
    assert func("magical", "blah") == "load blah"
    # Nothing should change
    assert len(due.citations) == 1
    assert len(due._entries) == 1
    assert due.citations[(path, "XXX0")].count == 2  # Besides the count

    # Now cause new citation given another value
    assert func("magical", "boo") == "load boo"
    assert len(due.citations) == 2
    assert len(due._entries) == 2
    assert (
        due.citations[(path, "XXX0")].count == 2
    )  # Count should stay the same for XXX0
    assert (
        due.citations[(path, "10.3389/fninf.2012.00022")].count == 1
    )  # but we get a new one


def test_dcite_match_conditions_function() -> None:
    due = DueCreditCollector()
    due.add(BibTeX(_sample_bibtex))

    @due.dcite(
        "XXX0", path="callable", conditions={(1, "kwarg2"): {"blah", "DC_DEFAULT"}}
    )
    @due.dcite(Doi(_sample_doi), path="callable", conditions={(1, "kwarg2"): {"boo"}})
    def method(arg1: str, kwarg2: str = "blah") -> str:
        """docstring"""
        assert arg1 == "magical"
        return "load %s" % kwarg2

    _test_dcite_match_conditions(due, method, "callable")


def test_dcite_match_conditions_method() -> None:
    due = DueCreditCollector()
    due.add(BibTeX(_sample_bibtex))

    class Citeable:
        def __init__(self, param: str | None = None) -> None:
            self.param = param

        @due.dcite(
            "XXX0",
            path="obj.callable",
            conditions={
                (2, "kwarg2"): {"blah", "DC_DEFAULT"},
                (0, "self.param"): {"paramvalue"},  # must be matched
            },
        )
        @due.dcite(
            Doi(_sample_doi), path="obj.callable", conditions={(2, "kwarg2"): {"boo"}}
        )
        def method(self, arg1: str, kwarg2: str = "blah") -> str:
            """docstring"""
            assert arg1 == "magical"
            return "load %s" % kwarg2

    citeable = Citeable(param="paramvalue")
    _test_dcite_match_conditions(due, citeable.method, "obj.callable")

    # now test for self.param -


def test_get_output_handler_method(tmpdir: py.path.local, monkeypatch) -> None:
    tempfile = str(tmpdir.mkdir("sub").join("tempfile.txt"))
    monkeypatch.setitem(os.environ, "DUECREDIT_OUTPUTS", "pickle")
    entry = BibTeX(_sample_bibtex)
    collector = DueCreditCollector()
    collector.cite(entry, path="module")

    summary = CollectorSummary(collector, fn=tempfile)
    handlers = [summary._get_output_handler(type_, collector) for type_ in ["pickle"]]

    # assert isinstance(handlers[0], TextOutput)
    assert isinstance(handlers[0], PickleOutput)

    pytest.raises(
        NotImplementedError, summary._get_output_handler, "nothing", collector
    )


def test_collectors_uniform_api() -> None:
    def get_api(objs) -> list[str]:
        return [
            x
            for x in sorted(sum((dir(obj) for obj in objs), []))
            if not x.startswith("_") or x in "__call__"
        ]

    assert get_api([DueCreditCollector, DueSwitch]) == get_api(
        [InactiveDueCreditCollector]
    )


def _test__docs__(method: Any) -> None:
    assert "entry:" in method.__doc__
    assert "tags: " in method.__doc__


def test__docs__() -> None:
    _test__docs__(DueCreditCollector.cite)
    _test__docs__(DueCreditCollector.dcite)
    _test__docs__(Citation.__init__)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Provides an adapter to switch between two (active, inactive) collectors
"""
from __future__ import annotations

import atexit
import os
from typing import TYPE_CHECKING, Any

from .log import lgr
from .utils import never_fail

if TYPE_CHECKING:
    from duecredit.collector import DueCreditCollector

    from .stub import InactiveDueCreditCollector


def _get_duecredit_enable() -> bool:
    env_enable = os.environ.get("DUECREDIT_ENABLE", "no")
    if not env_enable.lower() in ("0", "1", "yes", "no", "true", "false"):
        lgr.warning(
            "Misunderstood value %s for DUECREDIT_ENABLE. "
            "Use 'yes' or 'no', or '0' or '1'"
        )
    return env_enable.lower() in ("1", "yes", "true")


@never_fail
def _get_inactive_due() -> InactiveDueCreditCollector:
    from .stub import InactiveDueCreditCollector

    return InactiveDueCreditCollector()


@never_fail
def _get_active_due() -> DueCreditCollector | InactiveDueCreditCollector:
    from duecredit.collector import DueCreditCollector

    from .config import DUECREDIT_FILE
    from .io import load_due

    # TODO:  this needs to move to atexit handling, that we load previous
    # one and them merge with new ones.  Informative bits could be -- how
    # many new citations we got
    if os.path.exists(DUECREDIT_FILE):
        try:
            due_ = load_due(DUECREDIT_FILE)
        except Exception:
            lgr.warning(
                "Failed to load previously collected %s. "
                "DueCredit will not be active for this session." % DUECREDIT_FILE
            )
            return _get_inactive_due()
    else:
        due_ = DueCreditCollector()

    return due_


class DueSwitch:
    """Adapter between two types of collectors -- Inactive and Active

    Once activated though, cannot be fully deactivated since it would inject
    duecredit decorators and register an event atexit.
    """

    def __init__(self, inactive, active, activate: bool = False) -> None:
        self.__active: bool | None = None
        self.__collectors = {False: inactive, True: active}
        self.__activations_done = False
        if not (inactive and active):
            raise ValueError(
                "Both inactive and active collectors should be provided. "
                "Got active=%r, inactive=%r" % (active, inactive)
            )
        self.activate(activate)

    @property
    def active(self):
        return self.__active

    @never_fail
    def dump(self, **kwargs: Any) -> None:
        """Dumps summary of the citations

        Parameters
        ----------
        **kwargs: dict
           Passed to `CollectorSummary` constructor.
        """
        from duecredit.collector import CollectorSummary

        due_summary = CollectorSummary(self.__collectors[True], **kwargs)
        due_summary.dump()

    def __prepare_exit_and_injections(self) -> None:
        # Wrapper to create and dump summary... passing method doesn't work:
        #  probably removes instance too early

        atexit.register(self.dump)

        # Deal with injector
        from .injections import DueCreditInjector

        injector = DueCreditInjector(collector=self.__collectors[True])
        injector.activate()

    @never_fail
    def activate(self, activate: bool = True) -> None:
        # 1st step -- if activating/deactivating switch between the two collectors
        if self.__active is not activate:
            # we need to switch the state
            # InactiveDueCollector also has those special methods defined
            # in DueSwitch so that client code could query/call (for no effect).
            # So we shouldn't delete or bind them either
            def is_public_or_special(x):
                return not (x.startswith("_") or x in ("activate", "active", "dump"))

            # Clean up current bindings first
            for k in filter(is_public_or_special, dir(self)):
                delattr(self, k)

            new_due = self.__collectors[activate]
            for k in filter(is_public_or_special, dir(new_due)):
                setattr(self, k, getattr(new_due, k))

            self.__active = activate

        # 2nd -- if activating, we might still need to have activations done
        if activate and not self.__activations_done:
            try:
                self.__prepare_exit_and_injections()
            except Exception as e:
                lgr.error("Failed to prepare injections etc: %s" % str(e))
            finally:
                self.__activations_done = True


due = DueSwitch(_get_inactive_due(), _get_active_due(), _get_duecredit_enable())



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Helper to use duecredit as a "runnable" module with  -m duecredit"""
from __future__ import annotations

import sys

from . import __version__, due
from .log import lgr


def usage(outfile, executable=sys.argv[0]):
    if "__main__.py" in executable:
        # That was -m duecredit way to launch
        executable = "%s -m duecredit" % sys.executable
    outfile.write(
        """Usage: %s [OPTIONS] <file> [ARGS]

Meta-options:
--help                Display this help then exit.
--version             Output version information then exit.
"""
        % executable
    )


def runctx(cmd, global_ctx=None, local_ctx=None):
    if global_ctx is None:
        global_ctx = {}
    if local_ctx is None:
        local_ctx = {}

    try:
        exec(cmd, global_ctx, local_ctx)
    finally:
        # good opportunity to avoid atexit I guess. pass for now
        pass


def main(argv=None):
    import getopt
    import os

    if argv is None:
        argv = sys.argv

    try:
        opts, prog_argv = getopt.getopt(argv[1:], "", ["help", "version"])
        # TODO: support options for whatever we would support ;)
        # probably needs to hook in somehow into commands/options available
        # under cmdline/
    except getopt.error as msg:
        sys.stderr.write("{}: {}\n".format(sys.argv[0], msg))
        sys.stderr.write("Try `%s --help' for more information\n" % sys.argv[0])
        sys.exit(1)

    # and now we need to execute target script "manually"
    # Borrowing up on from trace.py
    for opt, _ in opts:
        if opt == "--help":
            usage(sys.stdout, executable=argv[0])
            sys.exit(0)

        if opt == "--version":
            sys.stdout.write("duecredit %s\n" % __version__)
            sys.exit(0)

    sys.argv = prog_argv
    progname = prog_argv[0]
    sys.path[0] = os.path.split(progname)[0]

    try:
        with open(progname) as fp:
            code = compile(fp.read(), progname, "exec")
        # try to emulate __main__ namespace as much as possible
        globs = {
            "__file__": progname,
            "__name__": "__main__",
            "__package__": None,
            "__cached__": None,
        }
        # Since used explicitly -- activate the beast
        due.activate(True)
        runctx(code, globs, globs)
        # TODO: see if we could hide our presence from the final tracebacks if execution fails
    except OSError as err:
        lgr.error("Cannot run file {!r} because: {}".format(sys.argv[0], err))
        sys.exit(1)
    except SystemExit:
        pass


if __name__ == "__main__":
    main()



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
from __future__ import annotations

# Just for testing of robust operation
import os

if "DUECREDIT_TEST_EARLY_IMPORT_ERROR" in os.environ.keys():
    raise ImportError("DUECREDIT_TEST_EARLY_IMPORT_ERROR")

from collections import defaultdict
import locale
from os.path import dirname, exists
import pickle
import re
import tempfile
import time
from typing import TYPE_CHECKING, Any
import warnings

from packaging.version import Version

from .config import CACHE_DIR, DUECREDIT_FILE
from .entries import BibTeX, Doi, DueCreditEntry, Text, Url
from .log import lgr
from .versions import external_versions

if TYPE_CHECKING:
    from .collector import Citation

_PREFERRED_ENCODING = locale.getpreferredencoding()


def get_doi_cache_file(doi: str) -> str:
    # where to cache bibtex entries
    if not os.path.exists(CACHE_DIR):
        os.makedirs(CACHE_DIR)
    return os.path.join(CACHE_DIR, doi)


def import_doi(doi: str, sleep: float = 0.5, retries: int = 10) -> str:
    import requests

    cached = get_doi_cache_file(doi)

    if exists(cached):
        with open(cached) as f:
            doi = f.read()
            return doi

    # else -- fetch it
    headers = {"Accept": "application/x-bibtex; charset=utf-8"}
    url = "https://doi.org/" + doi
    while retries > 0:
        lgr.debug("Submitting GET to %s with headers %s", url, headers)
        r = requests.get(url, headers=headers)
        r.encoding = "UTF-8"
        bibtex = r.text.strip()
        if bibtex.startswith("@"):
            # no more retries necessary
            break
        lgr.warning("Failed to obtain bibtex from doi.org, retrying...")
        time.sleep(sleep)  # give some time to the server
        retries -= 1
    status_code = r.status_code
    if not bibtex.startswith("@"):
        raise ValueError(
            "Query %(url)s for BibTeX for a DOI %(doi)s (wrong doi?) has failed. "
            "Response code %(status_code)d. "
            # 'BibTeX response was: %(bibtex)s'
            % locals()
        )
    if not exists(cached):
        cache_dir = dirname(cached)
        if not exists(cache_dir):
            os.makedirs(cache_dir)
        with open(cached, "w") as f:
            f.write(bibtex)
    return bibtex


def _is_contained(toppath: str, subpath: str) -> bool:
    if ":" not in toppath:
        return (
            (toppath == subpath)
            or (subpath.startswith(toppath + "."))
            or (subpath.startswith(toppath + ":"))
        )
    else:
        return subpath.startswith(toppath + ".")


class Output:
    """A generic class for setting up citations that then will be outputted
    differently (e.g., Bibtex, Text, etc.)"""

    def __init__(self, fd, collector) -> None:
        self.fd = fd
        self.collector = collector

    def _get_collated_citations(
        self, tags: list[str] | None = None, all_: bool | None = None
    ) -> tuple[
        dict[str, list[Citation]], dict[str, list[Citation]], dict[str, list[Citation]]
    ]:
        """Given all the citations, filter only those that the user wants and
        those that were actually used"""
        if not tags:
            env = os.environ.get(
                "DUECREDIT_REPORT_TAGS",
                "reference-implementation,implementation,dataset",
            )
            assert type(env) is str
            tags = env.split(",")
        if all_ is None:
            # consult env var
            env = os.environ.get("DUECREDIT_REPORT_ALL", "").lower()
            assert type(env) is str
            all_ = env in {"1", "true", "yes", "on"}
        tagset = set(tags)

        citations = self.collector.citations
        if tagset != {"*"}:
            # Filter out citations based on tags
            citations = {
                k: c for k, c in citations.items() if tagset.intersection(c.tags)
            }

        packages = defaultdict(list)
        modules = defaultdict(list)
        objects = defaultdict(list)

        # store the citations according to their path and divide them into
        # the right level
        for (path, _entry_key), citation in citations.items():
            if ":" in path:
                objects[path].append(citation)
            elif "." in path:
                modules[path].append(citation)
            else:
                packages[path].append(citation)

        # now we need to filter out the packages that don't have modules
        # or objects cited, or are specifically requested
        cited_packages = list(packages)
        cited_modobj = list(modules) + list(objects)
        for package in cited_packages:
            package_citations = packages[package]
            if (
                all_
                or any(
                    filter(lambda x: x.cite_module, package_citations)  # type: ignore
                )
                or any(
                    filter(lambda x: _is_contained(package, x), cited_modobj)  # type: ignore
                )
            ):
                continue
            else:
                # we don't need it
                del packages[package]

        return packages, modules, objects

    def dump(self, tags=None) -> None:
        raise NotImplementedError


class TextOutput(Output):
    def __init__(self, fd, collector, style=None) -> None:
        super().__init__(fd, collector)
        self.style = style
        if "DUECREDIT_STYLE" in os.environ.keys():
            self.style = os.environ["DUECREDIT_STYLE"]
        else:
            self.style = "harvard1"

    @staticmethod
    def _format_citations(citations, citation_nr) -> str:
        descriptions = map(str, {str(r.description) for r in citations})
        versions = map(str, {str(r.version) for r in citations})
        refnrs = map(str, [citation_nr[c.entry.key] for c in citations])
        path = citations[0].path

        return "- {} / {} (v {}) [{}]\n".format(
            ", ".join(descriptions), path, ", ".join(versions), ", ".join(refnrs)
        )

    def dump(self, tags=None) -> None:
        # get 'model' of citations
        packages, modules, objects = self._get_collated_citations(tags)
        # put everything into a single dict
        pmo = {}
        pmo.update(packages)
        pmo.update(modules)
        pmo.update(objects)

        # get all the paths
        paths = sorted(list(pmo))
        # get all the entry_keys in order
        entry_keys = [c.entry.key for p in paths for c in pmo[p]]
        # make a dictionary entry_key -> citation_nr
        citation_nr = defaultdict(int)
        refnr = 1
        for entry_key in entry_keys:
            if entry_key not in citation_nr:
                citation_nr[entry_key] = refnr
                refnr += 1

        self.fd.write("\nDueCredit Report:\n")
        start_refnr = 1
        for path in paths:
            # since they're lexicographically sorted by path, dependencies
            # should be maintained
            cites = pmo[path]
            if ":" in path or "." in path:
                self.fd.write("  ")
            self.fd.write(self._format_citations(cites, citation_nr))
            start_refnr += len(cites)

        # Print out some stats
        stats = [
            (len(packages), "package"),
            (len(modules), "module"),
            (len(objects), "function"),
        ]
        for n, cit_type in stats:
            self.fd.write(
                "\n{} {} cited".format(n, cit_type if n == 1 else cit_type + "s")
            )
        # now print out references
        printed_keys = []
        if len(pmo) > 0:
            self.fd.write("\n\nReferences\n" + "-" * 10 + "\n")
            for path in paths:
                for cit in pmo[path]:
                    # 'import Citation / assert type(cit) is Citation' would pollute environment
                    ek = cit.entry.key  # type: ignore
                    if ek not in printed_keys:
                        self.fd.write(f"\n[{citation_nr[ek]}] ")
                        self.fd.write(get_text_rendering(cit.entry, style=self.style))
                        printed_keys.append(ek)
            self.fd.write("\n")


def get_text_rendering(entry: DueCreditEntry, style: str = "harvard1") -> str:
    if isinstance(entry, Doi):
        return format_bibtex(get_bibtex_rendering(entry), style=style)
    elif isinstance(entry, BibTeX):
        return format_bibtex(entry, style=style)
    elif isinstance(entry, Text):
        return entry.format()
    elif isinstance(entry, Url):
        return f"URL: {entry.format()}"
    else:
        return str(entry)


def get_bibtex_rendering(entry: DueCreditEntry) -> BibTeX:
    if isinstance(entry, Doi):
        return BibTeX(import_doi(entry.doi))
    elif isinstance(entry, BibTeX):
        return entry
    else:
        raise ValueError("Have no clue how to get bibtex out of %s" % entry)


def condition_bibtex(bibtex: str) -> bytes:
    """Given a bibtex entry, "condition" it for processing with citeproc

    Primarily a set of workarounds for either non-standard BibTeX entries
    or citeproc bugs
    """
    # XXX: workaround atm to fix zenodo bibtexs, convert @data to @misc
    # and also ; into and
    if bibtex.startswith("@data"):
        bibtex = bibtex.replace("@data", "@misc", 1)
        bibtex = bibtex.replace(";", " and")
    bibtex = bibtex.replace("\u2013", "--") + "\n"
    # workaround for citeproc 0.3.0 failing to parse a single page pages field
    # as for BIDS paper.  Workaround to add trailing + after pages number
    # related issue asking for a new release: https://github.com/brechtm/citeproc-py/issues/72
    bibtex = re.sub(r'(pages\s*=\s*["{]\d+)(["}])', r"\1+\2", bibtex)
    # partial workaround for citeproc failing to parse page numbers when they contain non-numeric characters
    # remove opening letter, e.g. 'S123' -> '123'
    # related issue: https://github.com/brechtm/citeproc-py/issues/74
    bibtex = re.sub(r'(pages\s*=\s*["{])([a-zA-Z])', r"\g<1>", bibtex)
    return bibtex.encode("utf-8")


def format_bibtex(bibtex_entry: BibTeX, style: str = "harvard1") -> str:
    try:
        import citeproc as cp
        from citeproc.source.bibtex import BibTeX as cpBibTeX
    except ImportError as e:
        raise RuntimeError(
            "For formatted output we need citeproc and all of its dependencies "
            "(such as lxml) but there is a problem while importing citeproc: %s"
            % str(e)
        )
    decode_exceptions: tuple[type[Exception], ...]
    try:
        from citeproc.source.bibtex.bibparse import BibTeXDecodeError

        decode_exceptions = (UnicodeDecodeError, BibTeXDecodeError)
    except ImportError:
        # this version doesn't yet have this exception defined
        decode_exceptions = (UnicodeDecodeError,)
    key = bibtex_entry.get_key()
    # need to save it temporarily to use citeproc-py
    fname = tempfile.mktemp(suffix=".bib")
    try:
        with open(fname, "wb") as f:
            f.write(condition_bibtex(bibtex_entry.rawentry))
        # We need to avoid cpBibTex spitting out warnings
        old_filters = warnings.filters[:]  # store a copy of filters
        warnings.simplefilter("ignore", UserWarning)
        try:
            try:
                bib_source = cpBibTeX(fname)
            except decode_exceptions:
                # So .bib must be having UTF-8 characters.  With
                # a recent (not yet released past v0.3.0-68-g9800dad
                # we should be able to provide encoding argument
                bib_source = cpBibTeX(fname, encoding="utf-8")
        except Exception as e:
            msg = "Failed to process BibTeX file {}: {}.".format(fname, e)
            if "unexpected keyword argument" in str(e):
                citeproc_version = external_versions["citeproc"]
                if isinstance(citeproc_version, Version) and citeproc_version < Version(
                    "0.4"
                ):
                    err = "need a newer citeproc-py >= 0.4.0"
                    msg += " You might just " + err
            else:
                err = str(e)
            lgr.error(msg)
            return "ERRORED: %s" % err
        finally:
            # return warnings back
            warnings.filters = old_filters
        bib_style = cp.CitationStylesStyle(style, validate=False)
        # TODO: specify which tags of formatter we want
        bibliography = cp.CitationStylesBibliography(
            bib_style, bib_source, cp.formatter.plain
        )
        citation = cp.Citation([cp.CitationItem(key)])
        bibliography.register(citation)
    finally:
        if not os.environ.get("DUECREDIT_KEEPTEMP"):
            for i in range(50):
                try:
                    os.unlink(fname)
                except OSError:
                    if i < 49:
                        time.sleep(0.1)
                        continue
                    else:
                        raise
                break

    biblio_out = bibliography.bibliography()
    assert len(biblio_out) == 1
    biblio_out = "".join(biblio_out[0])
    return biblio_out  # if biblio_out else str(bibtex_entry)


# TODO: harmonize order of arguments
class PickleOutput:
    def __init__(self, collector, fn=DUECREDIT_FILE) -> None:
        self.collector = collector
        self.fn = fn

    def dump(self) -> None:
        with open(self.fn, "wb") as f:
            pickle.dump(self.collector, f)

    @classmethod
    def load(cls, filename: str = DUECREDIT_FILE) -> Any:
        with open(filename, "rb") as f:
            return pickle.load(f)


class BibTeXOutput(Output):
    def __init__(self, fd, collector) -> None:
        super().__init__(fd, collector)

    def dump(self, tags=None) -> None:
        packages, modules, objects = self._get_collated_citations(tags)
        # get all the citations in order
        pmo = {}
        pmo.update(packages)
        pmo.update(modules)
        pmo.update(objects)

        # get all the paths
        paths = sorted(list(pmo))

        entries = []
        for path in paths:
            for c in pmo[path]:
                if c.entry not in entries:
                    entries.append(c.entry)

        for entry in entries:
            try:
                bibtex = get_bibtex_rendering(entry)
            except Exception:
                lgr.warning("Failed to generate BibTeX for %s", entry)
                continue
            self.fd.write(bibtex.rawentry + "\n")


def load_due(filename: str) -> Any:
    return PickleOutput.load(filename)



# A tiny analysis script to demonstrate duecredit
#
# Import of duecredit is not necessary if you just run this script with
# python -m duecredit
# import duecredit  # Just to enable duecredit
from scipy.cluster.hierarchy import linkage
from scipy.spatial.distance import pdist
from sklearn.datasets import make_blobs

print("I: Simulating 4 blobs")
data, true_label = make_blobs(centers=4)

dist = pdist(data, metric="euclidean")

Z = linkage(dist, method="single")
print("I: Done clustering 4 blobs")



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Run internal DueCredit (unit)tests to verify correct operation on the system"""
from __future__ import annotations

__docformat__ = "restructuredtext"

import argparse

# magic line for manpage summary
# man: -*- % run unit-tests


def setup_parser(parser: argparse.ArgumentParser) -> None:
    # TODO -- pass options such as verbosity etc
    pass


def run(_args: argparse.Namespace) -> None:
    import duecredit  # noqa: F401

    raise NotImplementedError("Just use pytest duecredit for now")
    # duecredit.test()



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
Automatic injection of bibliography entries for nipype module
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..entries import BibTeX, Doi

# If defined, would determine from which to which version of the corresponding
# module to care about
min_version = None
max_version = None

if TYPE_CHECKING:
    from .injector import DueCreditInjector


def inject(injector: DueCreditInjector) -> None:
    # http://nipy.org/nipype/about.html
    injector.add(
        "nipype",
        None,
        Doi("10.3389/fninf.2011.00013"),
        description="Nipype: a flexible, lightweight and extensible neuroimaging data processing framework in Python",
        tags=["implementation"],
    )

    # http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/
    injector.add(
        "nipype.interfaces",
        "fsl",
        Doi("10.1016/j.neuroimage.2004.07.051"),
        description="Advances in functional and structural MR image analysis and implementation as FSL",
        tags=["implementation"],
    )
    injector.add(
        "nipype.interfaces",
        "fsl",
        Doi("10.1016/j.neuroimage.2008.10.055"),
        description="Bayesian analysis of neuroimaging data in FSL",
        tags=["implementation"],
    )
    injector.add(
        "nipype.interfaces",
        "fsl",
        Doi("10.1016/j.neuroimage.2011.09.015"),
        description="FSL.",
        tags=["implementation"],
    )

    # http://www.fil.ion.ucl.ac.uk/spm
    injector.add(
        "nipype.interfaces",
        "spm",
        BibTeX(
            """
        @book{FrackowiakFristonFrithDolanMazziotta1997,
            author={R.S.J. Frackowiak, K.J. Friston, C.D. Frith, R.J. Dolan, and J.C. Mazziotta},
            title={Human Brain Function},
            publisher={Academic Press USA}
            year={1997},
            }
        """
        ),
        description="The fundamental text on Statistical Parametric Mapping (SPM)",
        tags=["implementation"],
    )

    # http://surfer.nmr.mgh.harvard.edu/fswiki/FreeSurferMethodsCitation
    # there are a bunch, not sure what is primary
    # injector.add('nipype.interfaces', 'freesurfer', Doi(''),
    #                description='',
    #                tags=['implementation'])

    # http://afni.nimh.nih.gov/afni/about/citations/
    # there are a bunch, not sure what is primary
    # injector.add('nipype.interfaces', 'afni', Doi(''),
    #                description='',
    #                tags=['implementation'])



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Automatic injection of bibliography entries for nipy module
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..entries import Doi

# If defined, would determine from which to which version of the corresponding
# module to care about
min_version = None
max_version = None

if TYPE_CHECKING:
    from .injector import DueCreditInjector


def inject(injector: DueCreditInjector) -> None:
    injector.add(
        "nipy",
        None,
        Doi("10.1016/S1053-8119(09)72223-2"),
        description="Library fMRI data analysis",
        tags=["implementation"],
    )

    for f, d in [
        (
            "spectral_decomposition",
            "PCA decomposition of symbolic HRF shifted over time",
        ),
        ("taylor_approx", "A Taylor series approximation of an HRF shifted over time"),
    ]:
        injector.add(
            "nipy.modalities.fmri.fmristat.hrf",
            f,
            Doi("10.1006/nimg.2002.1096"),
            description=d,
            tags=["implementation"],
        )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.   Originates from datalad package distributed
#   under MIT license
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
from __future__ import annotations

import logging
import re

lgr = logging.getLogger("duecredit.entries")


class DueCreditEntry:
    _key: str

    def __init__(self, rawentry: str, key: str | None = None) -> None:
        self._rawentry = rawentry
        self._key = key or rawentry.lower()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DueCreditEntry):
            return NotImplemented
        return (self._rawentry == other._rawentry) and (self._key == other._key)

    def get_key(self) -> str:
        return self._key

    @property
    def key(self) -> str:
        return self.get_key()

    @property
    def rawentry(self) -> str:
        return self._rawentry

    def _process_rawentry(self) -> None:
        pass

    def __repr__(self) -> str:
        argl = [repr(self._rawentry), f"key={repr(self._key)}"]
        args = ", ".join(argl)
        return self.__class__.__name__ + f"({args})"

    def format(self) -> str:
        # TODO: return nice formatting of the entry
        return str(self._rawentry)


class BibTeX(DueCreditEntry):
    def __init__(self, bibtex: str, key: str | None = None) -> None:
        super().__init__(bibtex.strip())
        self._reference = None
        self._process_rawentry()
        if key is not None:
            # use the one provided, not the parsed one
            lgr.debug(
                "Replacing parsed key %s for BibTeX with the provided %s",
                self._key,
                key,
            )
            self._key = key

    def _process_rawentry(self) -> None:
        reg = re.match(
            r"\s*@(?P<type>\S*)\s*\{\s*(?P<key>\S*)\s*,.*",
            self._rawentry,
            flags=re.MULTILINE,
        )
        assert reg
        matches = reg.groupdict()
        self._key = matches["key"]


class Text(DueCreditEntry):
    """Just a free text entry without any special super powers in rendering etc"""

    pass  # nothing special I guess


class Doi(DueCreditEntry):
    @property
    def doi(self) -> str:
        return self._rawentry


class Url(DueCreditEntry):
    @property
    def url(self) -> str:
        return self._rawentry



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Automatic injection of bibliography entries for numpy module
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..entries import BibTeX

# If defined, would determine from which to which version of the corresponding
# module to care about
min_version = None
max_version = None

if TYPE_CHECKING:
    from .injector import DueCreditInjector


def inject(injector: DueCreditInjector) -> None:
    injector.add(
        "numpy",
        None,
        BibTeX(
            r"""
    @article{van2011numpy,
        title={The NumPy array: a structure for efficient numerical computation},
        author={Van Der Walt, Stefan and Colbert, S Chris and Varoquaux, Gael},
        journal={Computing in Science \& Engineering},
        volume={13},
        number={2},
        pages={22--30},
        year={2011},
        publisher={AIP Publishing},
        doi={10.1109/MCSE.2011.37}
        }
    """
        ),
        tags=["implementation"],
        cite_module=True,
        description="Scientific tools library",
    )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
from __future__ import annotations

from os import linesep

import pytest

from ..version import __version__
from ..versions import ExternalVersions, Version


# just to ease testing
def cmp(a, b):
    return (a > b) - (a < b)


def test_external_versions_basic():
    ev = ExternalVersions()
    assert ev._versions == {}
    assert ev["duecredit"] == Version(__version__)
    # and it could be compared
    assert ev["duecredit"] >= Version(__version__)
    assert ev["duecredit"] > Version("0.1")
    assert list(ev.keys()) == ["duecredit"]
    assert "duecredit" in ev
    assert "unknown" not in ev

    # Version might remove training .0
    version_str = (
        str(ev["duecredit"]) if isinstance(ev["duecredit"], Version) else __version__
    )
    assert ev.dumps() == "Versions: duecredit=%s" % version_str

    # For non-existing one we get None
    assert ev["duecreditnonexisting"] is None

    # and nothing gets added to _versions for nonexisting
    assert set(ev._versions.keys()) == {"duecredit"}

    # but if it is a module without version, we get it set to UNKNOWN
    assert ev["os"] == ev.UNKNOWN
    # And get a record on that inside
    assert ev._versions.get("os") == ev.UNKNOWN
    # And that thing is "True", i.e. present
    assert ev["os"]
    # but not comparable with anything besides itself (was above)
    with pytest.raises(TypeError):
        cmp(ev["os"], "0")

    # assert_raises(TypeError, assert_greater, ev['os'], '0')

    # And we can get versions based on modules themselves
    from duecredit.tests import mod

    assert ev[mod] == Version(mod.__version__)

    # Check that we can get a copy of the versions
    versions_dict = ev.versions
    versions_dict["duecredit"] = "0.0.1"
    assert versions_dict["duecredit"] == "0.0.1"
    assert ev["duecredit"] == Version(__version__)


def test_external_versions_unknown() -> None:
    assert str(ExternalVersions.UNKNOWN) == "UNKNOWN"


def _test_external(ev, modname: str) -> None:
    try:
        exec("import %s" % modname, globals(), locals())
    except ImportError:
        modname = pytest.importorskip(modname)
    except Exception as e:
        pytest.skip("External {} fails to import: {}".format(modname, e))
    assert ev[modname] is not ev.UNKNOWN
    assert ev[modname] > Version("0.0.1")
    assert Version("1000000.0") > ev[modname]  # unlikely in our lifetimes


@pytest.mark.parametrize(
    "modname",
    [
        "scipy",
        "numpy",
        "mvpa2",
        "sklearn",
        "statsmodels",
        "pandas",
        "matplotlib",
        "psychopy",
    ],
)
def test_external_versions_popular_packages(modname: str) -> None:
    ev = ExternalVersions()

    _test_external(ev, modname)

    # more of a smoke test
    assert linesep not in ev.dumps()
    assert ev.dumps(indent=True).endswith(linesep)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Spit out the summary of the citations which were collected.

"""
from __future__ import annotations

import argparse
import os
import sys

from ..config import DUECREDIT_FILE
from ..io import BibTeXOutput, TextOutput
from ..log import lgr

__docformat__ = "restructuredtext"

# magic line for manpage summary
# man: -*- % summary of collected citations


def setup_parser(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-f",
        "--filename",
        default=DUECREDIT_FILE,
        help="Filename containing collected citations. Default: %(default)s",
    )

    parser.add_argument(
        "--style",
        choices=("apa", "harvard1"),
        default="harvard1",
        help="Style to be used for listing citations",
    )

    parser.add_argument(
        "--format",
        choices=("text", "bibtex"),
        default="text",
        help="Way to present the summary",
    )


def run(args: argparse.Namespace) -> int:
    from ..io import PickleOutput

    if not os.path.exists(args.filename):
        lgr.debug("File {} doesn't exist.  No summary available".format(args.filename))
        return 1

    due = PickleOutput.load(args.filename)
    # CollectorSummary(due).dump()

    out: TextOutput | BibTeXOutput
    if args.format == "text":
        out = TextOutput(sys.stdout, due, args.style)
    elif args.format == "bibtex":
        out = BibTeXOutput(sys.stdout, due)
    else:
        raise ValueError("unknown to treat %s" % args.format)
    out.dump()
    return 0



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
"""

__docformat__ = "restructuredtext"

# argument spec template
# <name> = (
#    <id_as_positional>, <id_as_option>
#    {<ArgusmentParser.add_arguments_kwargs>}
# )

from ..cmdline.helpers import HelpAction, LogLevelAction

help = (  # noqa: A001
    "help",
    ("-h", "--help", "--help-np"),
    dict(
        nargs=0,
        action=HelpAction,
        help="""show this help message and exit. --help-np forcefully disables
                 the use of a pager for displaying the help.""",
    ),
)

version = (
    "version",
    ("--version",),
    dict(
        action="version", help="show program's version and license information and exit"
    ),
)

log_level = (
    "log-level",
    ("-l", "--log-level"),
    dict(
        action=LogLevelAction,
        choices=["critical", "error", "warning", "info", "debug"]
        + [str(x) for x in range(1, 10)],
        default="warning",
        help="""level of verbosity. Integers provide even more debugging information""",
    ),
)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

"""
Automatic injection of bibliography entries for mdp module
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..entries import BibTeX, Doi

# If defined, would determine from which to which version of the corresponding
# module to care about
min_version = None
max_version = None

if TYPE_CHECKING:
    from .injector import DueCreditInjector


def inject(injector: DueCreditInjector) -> None:
    injector.add(
        "mdp",
        None,
        Doi("10.3389/neuro.11.008.2008"),
        description="Modular toolkit for Data Processing (MDP): a Python data processing framework",
        tags=["implementation"],
    )

    injector.add(
        "mdp.nodes",
        "PCANode.train",
        Doi("10.1007/b98835"),
        description="Principal Component Analysis (and filtering)",
        tags=["implementation"],
    )

    injector.add(
        "mdp.nodes",
        "NIPALSNode.train",
        BibTeX(
            """
        @incollection{Word1966,
            author={Wold, H.},
            title={Nonlinear estimation by iterative least squares procedures.},
            booktitle={Research Papers in Statistics},
            publisher={Wiley}
            year={1966},
            editor={David, F.},
            pages={411--444},
            }
        """
        ),
        description="Principal Component Analysis using the NIPALS algorithm.",
        tags=["edu"],
    )

    injector.add(
        "mdp.nodes",
        "FastICANode.train",
        Doi("10.1109/72.761722"),
        description="Independent Component Analysis using the FastICA algorithm",
        tags=["implementation"],
    )

    injector.add(
        "mdp.nodes",
        "CuBICANode.train",
        Doi("10.1109/TSP.2004.826173"),
        description="Independent Component Analysis using the CuBICA algorithm.",
        tags=["implementation"],
    )

    injector.add(
        "mdp.nodes",
        "NIPALSNode.train",
        BibTeX(
            """
        @conference{ZieheMuller1998,
            author={Ziehe, Andreas and Muller, Klaus-Robert},
            title={TDSEP an efficient algorithm for blind separation using time structure.},
            booktitle={Proc. 8th Int. Conf. Artificial Neural Networks},
            year={1998},
            editor={Niklasson, L, Boden, M, and Ziemke, T},
            publisher={ICANN}
            }
        """
        ),
        description="Independent Component Analysis using the TDSEP algorithm",
        tags=["edu"],
    )

    injector.add(
        "mdp.nodes",
        "JADENode.train",
        Doi("10.1049/ip-f-2.1993.0054"),
        description="Independent Component Analysis using the JADE algorithm",
        tags=["implementation"],
    )
    injector.add(
        "mdp.nodes",
        "JADENode.train",
        Doi("10.1162/089976699300016863"),
        description="Independent Component Analysis using the JADE algorithm",
        tags=["implementation"],
    )

    injector.add(
        "mdp.nodes",
        "SFANode.train",
        Doi("10.1162/089976602317318938"),
        description="Slow Feature Analysis",
        tags=["implementation"],
    )

    injector.add(
        "mdp.nodes",
        "SFA2Node.train",
        Doi("10.1162/089976602317318938"),
        description="Slow Feature Analysis (via the space of inhomogeneous polynomials)",
        tags=["implementation"],
    )

    injector.add(
        "mdp.nodes",
        "ISFANode.train",
        Doi("10.1007/978-3-540-30110-3_94"),
        description="Independent Slow Feature Analysis",
        tags=["implementation"],
    )

    injector.add(
        "mdp.nodes",
        "XSFANode.train",
        BibTeX(
            """
        @article{SprekelerZitoWiskott2009,
            author={Sprekeler, H., Zito, T., and Wiskott, L.},
            title={An Extension of Slow Feature Analysis for Nonlinear Blind Source Separation.},
            journal={Journal of Machine Learning Research.},
            year={2009},
            volume={15},
            pages={921--947},
            }
        """
        ),
        description="Non-linear Blind Source Separation using Slow Feature Analysis",
        tags=["edu"],
    )

    injector.add(
        "mdp.nodes",
        "FDANode.train",
        BibTeX(
            """
        @book{Bishop2011,
            author={Bishop, Christopher M.},
            title={Neural Networks for Pattern Recognition},
            publisher={Oxford University Press, Inc}
            year={2011},
            pages={105--112},
            }
        """
        ),
        description="(generalized) Fisher Discriminant Analysis",
        tags=["edu"],
    )

    # etc...



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Automatic injection of bibliography entries for numpy module
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..entries import BibTeX, Doi, Url

# If defined, would determine from which to which version of the corresponding
# module to care about
min_version = None
max_version = None

if TYPE_CHECKING:
    from .injector import DueCreditInjector


def inject(injector: DueCreditInjector) -> None:
    injector.add(
        "sklearn",
        None,
        BibTeX(
            """
    @article{pedregosa2011scikit,
        title={Scikit-learn: Machine learning in Python},
        author={Pedregosa, Fabian and Varoquaux, Ga{\"e}l and Gramfort,
        Alexandre and Michel, Vincent and Thirion, Bertrand and Grisel,
        Olivier and Blondel, Mathieu and Prettenhofer, Peter and Weiss,
        Ron and Dubourg, Vincent and others},
        journal={The Journal of Machine Learning Research},
        volume={12},
        pages={2825--2830},
        year={2011},
        publisher={JMLR.org}
        }
    """
        ),
        description="Machine Learning library",
    )

    # sklearn.cluster.affinity_propagation_
    injector.add(
        "sklearn.cluster.affinity_propagation_",
        None,
        Doi("10.1126/science.1136800"),
        description="Affinity propagation clustering algorithm",
        tags=["implementation"],
    )

    # sklearn.cluster.bicluster
    injector.add(
        "sklearn.cluster.bicluster",
        "SpectralCoclustering._fit",
        Doi("10.1101/gr.648603"),
        description="Spectral Coclustering algorithm",
        tags=["implementation"],
    )
    injector.add(
        "sklearn.cluster.bicluster",
        "SpectralBiclustering._fit",
        Doi("10.1101/gr.648603"),
        description="Spectral Biclustering algorithm",
        tags=["implementation"],
    )

    # sklearn.cluster.birch
    injector.add(
        "sklearn.cluster.birch",
        "Birch._fit",
        Doi("10.1145/233269.233324"),
        description="BIRCH clustering algorithm",
        tags=["implementation"],
    )
    injector.add(
        "sklearn.cluster.birch",
        "Birch._fit",
        Url("https://code.google.com/p/jbirch/"),
        description="Java implementation of BIRCH clustering algorithm",
        tags=["another-implementation"],
    )

    # sklearn.cluster.dbscan_
    injector.add(
        "sklearn.cluster.dbscan_",
        "dbscan",
        BibTeX(
            """@inproceedings{ester1996density,
  title={A density-based algorithm for discovering clusters in large spatial databases with noise.},
  author={Ester, Martin and Kriegel, Hans-Peter and Sander, J{\"o}rg and Xu, Xiaowei},
  booktitle={Kdd},
  volume={96},
  number={34},
  pages={226--231},
  year={1996}
}"""
        ),
        description="dbscan clustering algorithm",
        tags=["implementation"],
    )

    # sklearn.cluster.mean_shift_
    injector.add(
        "sklearn.cluster.mean_shift_",
        "mean_shift",
        Doi("10.1109/34.1000236"),
        description="Mean shift clustering algorithm",
        tags=["implementation"],
    )

    # sklearn.cluster.spectral
    injector.add(
        "sklearn.cluster.spectral",
        "discretize",
        Doi("10.1109/ICCV.2003.1238361"),
        description="Multiclass spectral clustering",
        tags=["reference"],
    )
    injector.add(
        "sklearn.cluster.spectral",
        "spectral_clustering",
        Doi("10.1109/34.868688"),
        description="Spectral clustering",
        tags=["implementation"],
    )
    injector.add(
        "sklearn.cluster.spectral",
        "spectral_clustering",
        Doi("10.1007/s11222-007-9033-z"),
        description="Spectral clustering",
        tags=["implementation"],
    )

    # sklearn.ensemble.forest and tree
    Breiman_2001 = Doi("10.1023/A:1010933404324")
    Breiman_1984 = BibTeX(
        """@BOOK{breiman-friedman-olshen-stone-1984,
  author        = {L. Breiman and J. Friedman and R. Olshen and C. Stone},
  title         = {{Classification and Regression Trees}},
  publisher     = {Wadsworth and Brooks},
  address       = {Monterey, CA},
  year          = {1984},
}"""
    )
    # Not clear here though if those are the original publication on the topic
    # or just an educational references (books), most probably both ;)
    injector.add(
        "sklearn.ensemble.forest",
        "RandomForestClassifier.predict_proba",
        Breiman_2001,
        description="Random forest classifiers",
        tags=["implementation", "edu"],
    )
    injector.add(
        "sklearn.ensemble.forest",
        "RandomForestRegressor.predict",
        Breiman_2001,
        description="Random forest regressions",
        tags=["implementation", "edu"],
    )
    injector.add(
        "sklearn.tree.tree",
        "DecisionTreeClassifier.predict_proba",
        Breiman_1984,
        description="Classification and regression trees",
        tags=["implementation", "edu"],
    )
